# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Provides a set of functions to handle resampling.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import warnings
from IPython import embed
import numpy
from scipy import interpolate
import astropy.constants

from .covariance import Covariance

# TODO: spectral_coordinate_step, spectrum_velocity_scale, and
# angstroms_per_pixel should all use centers_to_borders

def spectral_coordinate_step(wave, log=False, base=10.0):
    """
    Return the *uniform* sampling step for the input wavelength
    vector.

    If the sampling is logarithmic, return the change in the
    logarithm of the wavelength; otherwise, return the linear step in
    angstroms.

    Args: 
        wave (`numpy.ndarray`_):
            Wavelength coordinates of each spectral channel in
            angstroms.
        log (:obj:`bool`, optional):
            Input spectrum has been sampled geometrically.
        base (:obj:`float`, optional):
            If sampled geometrically, the sampling is done using a
            logarithm with this base. For natural logarithm, use
            numpy.exp(1).

    Returns:
        :obj:`float`: Spectral sampling step in either angstroms
        (log=False) or the step in log(angstroms).

    Raises:
        ValueError:
            Raised if the wavelength vector is not linearly or
            log-linearly sampled.
    """
    dw = numpy.diff(numpy.log(wave))/numpy.log(base) if log else numpy.diff(wave)
    if numpy.any( numpy.absolute(numpy.diff(dw)) > 100*numpy.finfo(dw.dtype).eps):
        raise ValueError('Wavelength vector is not uniformly sampled to numerical accuracy.')
    return numpy.mean(dw)


def spectrum_velocity_scale(wave):
    """
    Determine the velocity sampling of an input wavelength vector when log sampled
    
    .. note::
        The wavelength vector is assumed to be geometrically sampled!
        However, the input units expected to be in angstroms, not, e.g.,
        log(angstrom).

    Args: 
        wave (numpy.ndarray): Wavelength coordinates of each spectral
            channel in angstroms.  It is expected that the spectrum has
            been sampled geometrically

    Returns:
        float: Velocity scale of the spectrum in km/s.

    """
    return astropy.constants.c.to('km/s').value*spectral_coordinate_step(wave, log=True,
                                                                         base=numpy.exp(1.))


def angstroms_per_pixel(wave, log=False, base=10.0, regular=True):
    """
    Return a vector with the angstroms per pixel at each channel.

    When `regular=True`, the function assumes that the wavelengths are
    either sampled linearly or geometrically.  Otherwise, it calculates
    the size of each pixel as the difference between the wavelength
    coordinates.  The first and last pixels are assumed to have a width
    as determined by assuming the coordinate is at its center.

    .. note::

        If the regular is False and log is True, the code does *not*
        assume the wavelength coordinates are at the geometric center of
        the pixel.

    Args:
        wave (`numpy.ndarray`_):
            (Geometric) centers of the spectrum pixels in angstroms.
        log (`numpy.ndarray`_, optional):
            The vector is geometrically sampled.
        base (:obj:`float`, optional):
            Base of the logarithm used in the geometric sampling.
        regular (:obj:`bool`, optional):
            The vector is regularly sampled.

    Returns:
        numpy.ndarray: The angstroms per pixel.
    """
    if regular:
        ang_per_pix = spectral_coordinate_step(wave, log=log, base=base)
        return ang_per_pix*wave*numpy.log(base) if log else numpy.repeat(ang_per_pix, len(wave))

    return numpy.diff([(3*wave[0]-wave[1])/2] + ((wave[1:] + wave[:-1])/2).tolist()
                      + [(3*wave[-1]-wave[-2])/2])


def grid_npix(rng=None, dx=None, log=False, base=10.0, default=None):
    """
    Determine the number of pixels needed for a given grid.

    Args:
        rng (array-like, optional):
            Two-element array with the
            starting and ending x coordinate of the pixel centers to
            divide into pixels of a given width.  If *log* is True, this
            must still be the linear value of the x coordinate, not
            log(x)!.
        dx (:obj:`float`, optional):
            Linear or logarithmic pixel width.
        log (:obj:`bool`, optional):
            Flag that the range should be logarithmically binned.
        base (:obj:`float`, optional):
            Base for the logarithm
        default (:obj:`int`, optional):
            Default number of pixels to use. The default is returned
            if either ``rng`` or ``dx`` are not provided.

    Returns:
        :obj:`tuple`: Returns the number of pixels to cover ``rng``
        with pixels of width ``dx`` and a two-element
        `numpy.ndarray`_ with the adjusted range such that number of
        pixels of size dx is the exact integer.

    Raises:
        ValueError:
            Raised if the range is not a two-element vector.
    """
    # If the range or sampling are not provided, the number of pixels is
    # already set
    if rng is None or dx is None:
        return default, rng
    if len(rng) != 2:
        raise ValueError('Range must be a 2-element vector.')

    _rng = numpy.atleast_1d(rng).copy()
    npix = int(numpy.floor(numpy.diff(numpy.log(_rng))/numpy.log(base)/dx) + 1) if log else \
                    int(numpy.floor(numpy.diff(_rng)/dx) + 1)
    _rng[1] = numpy.power(base, numpy.log(_rng[0])/numpy.log(base) + dx*(npix-1)) \
                            if log else _rng[0] + dx*(npix-1)

    # Fix for numerical precision
    if (not log and numpy.isclose(rng[1] - _rng[1], dx)) \
           or (log and numpy.isclose((numpy.log(rng[1]) - numpy.log(_rng[1]))/numpy.log(base), dx)):
        npix += 1
        _rng[1] = numpy.power(base, numpy.log(_rng[0])/numpy.log(base) + dx*(npix-1)) \
                                if log else _rng[0] + dx*(npix-1)

    return npix, _rng


def grid_borders(rng, npix, log=False, base=10.0):
    """
    Determine the borders of bin edges in a grid.

    Args:
        rng (array-like):
            Two-element array with the (geometric) centers of the
            first and last pixel in the grid.
        npix (:obj:`int`):
            Number of pixels in the grid.
        log (:obj:`bool`, optional):
            The input range is (to be) logarithmically sampled.
        base (:obj:`float`, optional):
            The base of the logarithmic sampling. Use
            ``numpy.exp(1.)`` for the natural logarithm.

    Returns:
        :obj:`tuple`: Returns a `numpy.ndarray`_ with the grid
        borders with shape ``(npix+1,)`` and the step size per grid
        point. If ``log=True``, the latter is the geometric step.
    """
    if log:
        _rng = numpy.log(rng)/numpy.log(base)
        dlogx = numpy.diff(_rng)[0]/(npix-1.)
        borders = numpy.power(base, numpy.linspace(*(_rng/dlogx + [-0.5, 0.5]), num=npix+1)*dlogx)
        return borders, dlogx
    dx = numpy.diff(rng)[0]/(npix-1.)
    borders = numpy.linspace(*(numpy.atleast_1d(rng)/dx + numpy.array([-0.5, 0.5])), num=npix+1)*dx
    return borders, dx


def grid_centers(rng, npix, log=False, base=10.0):
    """
    Determine the (geometric) center of pixels in a grid.

    Args:
        rng (array-like):
            Two-element array with the (geometric) centers of the
            first and last pixel in the grid.
        npix (:obj:`int`):
            Number of pixels in the grid.
        log (:obj:`bool`, optional):
            The input range is (to be) logarithmically sampled.
        base (:obj:`float`, optional):
            The base of the logarithmic sampling. Use
            ``numpy.exp(1.)`` for the natural logarithm.

    Returns:
        :obj:`tuple`: Returns a `numpy.ndarray`_ with the grid pixel
        (geometric) ceners with shape ``(npix,)`` and the step size
        per grid point. If ``log=True``, the latter is the geometric
        step.
    """
    if log:
        _rng = numpy.log(rng)/numpy.log(base)
        dlogx = numpy.diff(_rng)[0]/(npix-1.)
        centers = numpy.power(base, numpy.linspace(*(_rng/dlogx), num=npix)*dlogx)
        return centers, dlogx
    dx = numpy.diff(rng)[0]/(npix-1.)
    centers = numpy.linspace(*(numpy.atleast_1d(rng)/dx), num=npix)*dx
    return centers, dx


def borders_to_centers(borders, log=False):
    """
    Convert a set of bin borders to bin centers.

    Grid borders need not be regularly spaced.

    Args:
        borders (`numpy.ndarray`_):
            Borders for adjoining bins.
        log (:obj:`bool`, optional):
            Return the geometric center instead of the linear center
            of the bins.

    Returns:
        `numpy.ndarray`_: The vector of bin centers.
    """
    return numpy.sqrt(borders[:-1]*borders[1:]) if log else (borders[:-1]+borders[1:])/2.0


def centers_to_borders(x, log=False):
    """
    Convert a set of bin centers to bounding edges.

    Grid centers need not be regularly spaced. The first edge of the
    first bin and the last edge of the last bin are assumed to be
    equidistant from the center of the 2nd and penultimate bins,
    respectively.

    Args:
        x (`numpy.ndarray`_):
            Centers of adjoining bins.
        log (:obj:`bool`, optional):
            Adopt a geometric binning instead of a linear binning.

    Returns:
        `numpy.ndarray`_: The vector with the coordinates of
        adjoining bin edges.
    """
    if log:
        dx = numpy.diff(numpy.log(x))
        return numpy.exp(numpy.append(numpy.log(x[:-1]) - dx/2,
                                      numpy.log(x[-1]) + numpy.array([-1,1])*dx[-1]/2))
    dx = numpy.diff(x)
    return numpy.append(x[:-1] - dx/2, x[-1] + numpy.array([-1,1])*dx[-1]/2)


class Resample:
    r"""
    Resample regularly or irregularly sampled data to a new grid using
    integration.
    
    This is a generalization of the routine
    :func:`ppxf.ppxf_util.log_rebin` provided by Michele Cappellari in
    the pPXF package.

    The abscissa coordinates (``x``) or the pixel borders
    (``xBorders``) for the data (``y``) should be provided for
    irregularly sampled data. If the input data is linearly or
    geometrically sampled (``inLog=True``), the abscissa coordinates
    can be generated using the input range for the (geometric) center
    of each grid point. If ``x``, ``xBorders``, and ``xRange`` are
    all None, the function assumes grid coordinates of ``x =
    numpy.arange(y.shape[-1])``.

    The function resamples the data by constructing the borders of
    the output grid using the ``new*`` keywords and integrating the
    input function between those borders. The output data will be set
    to ``ext_value`` for any data beyond the abscissa limits of the
    input data.

    The data to resample (``y``) can be a 1D or 2D array; the
    abscissa coordinates must always be 1D. If ``y`` is 2D, the
    resampling is performed along the last axis (i.e., ``axis=-1``).

    The nominal assumption is that the provided function is a step
    function based on the provided input (i.e., ``step=True``). If
    the output grid is substantially finer than the input grid, the
    assumption of a step function will be very apparent. To assume
    the function is instead linearly interpolated between each
    provided point, choose ``step=False``; higher-order
    interpolations are not provided.

    If errors are provided, a nominal error propagation is performed
    to provide the errors in the resampled data.

    .. warning::

        Depending on the details of the resampling, the output errors
        are likely highly correlated.  Any later analysis of the
        resampled function should account for this.

    The covariance in the resampled pixels can be constructed by
    setting ``covar=True``; however, this is currently only supported
    when ``step=True``. If no errors are provided and ``covar=True``,
    the computed matrix is the *correlation* matrix instead of the
    *covariance* matrix. Given that the resampling is the same for all
    vectors, only one correlation matix will be calculated if no
    errors are provided, even if the input ``y`` is 2D. If the input
    data to be resampled is 2D and errors *are* provided, a
    covariance matrix is calculated for *each* vector in ``y``.
    Beware that this can be an expensive computation.

    The ``conserve`` keyword sets how the units of the input data
    should be treated. If ``conserve=False``, the input data are
    expected to be in density units (i.e., per ``x`` coordinate unit)
    such that the integral over :math:`dx` is independent of the
    units of :math:`x` (i.e., flux per unit angstrom or flux
    density). If ``conserve=True``, the value of the data is assumed
    to have been integrated over the size of each pixel (i.e., units
    of flux). If ``conserve=True``, :math:`y` is converted to units
    of per step in :math:`x` such that the integral before and after
    the resample is the same. For example, if :math:`y` is a spectrum
    in units of flux, the function first converts the units to flux
    density and then computes the integral over each new pixel to
    produce the new spectra with units of flux.

    .. todo::
        - Allow for higher order interpolations.
        - Enable covariance matrix calculations for ``step=False``.
        - Provide examples

    Args:
        y (`numpy.ndarray`_, `numpy.ma.MaskedArray`_):
            Data values to resample. The shape can be 1D or 2D. If
            1D, the shape must be :math:`(N_{\rm pix},)`; otherwise,
            it must be :math:`(N_y,N_{\rm pix})`. I.e., the length of
            the last axis must match the input coordinates.
        e (`numpy.ndarray`_, `numpy.ma.MaskedArray`_, optional):
            Errors in the data that should be resampled. The shape
            must match the input ``y`` array. These data are used to
            perform a nominal calculation of the error in the
            resampled array.
        mask (`numpy.ndarray`_, optional):
            A boolean array indicating values in ``y`` that should be
            ignored during the resampling (values to ignore have
            ``masked=True``, just like in a `numpy.ma.MaskedArray`_).
            The mask used during the resampling is the union of this
            object and the masks of ``y`` and ``e``, if either are
            provided as `numpy.ma.MaskedArray`_ objects.
        x (`numpy.ndarray`_, optional):
            Abscissa coordinates for the data, which do not need to
            be regularly sampled. If the pixel borders are not
            provided, they are assumed to be half-way between
            adjacent pixels, and the first and last borders are
            assumed to be equidistant about the provided value. If
            these coordinates are not provided, they are determined
            by the input borders, the input range, or just assumed to
            be the indices, :math:`0..N_{\rm pix}-1`.
        xRange (array-like, optional):
            A two-element array with the starting and ending value
            for the coordinates of the centers of the first and last
            pixels in ``y``. Default is :math:`[0,N_{\rm pix}-1]`.
        xBorders (`numpy.ndarray`_, optional):
            An array with the borders of each pixel that must have a
            length of :math:`N_{\rm pix}+1`.
        inLog (:obj:`bool`, optional):
            Flag that the input is logarithmically binned, primarily
            meaning that the coordinates are at the geometric center
            of each pixel and the centers are spaced logarithmically.
            If false, the sampling is expected to be linear.
        newx (array-like, optional):
            Abscissa coordinates for the *output* data, which do not
            need to be a regular grid. If this is provided, the pixel
            borders are assumed to be half-way between adjacent
            pixels, and the first and last borders are assumed to be
            equidistant about the provided value. If these
            coordinates are not provided, they are determined by the
            new range, the new number of pixels, and/or the new pixel
            width (and whether or not the new grid should be
            logarithmically binned). If this is provided,
            ``newRange``, ``newpix``, ``newLog``, and ``newdx`` are
            *all* ignored.
        newRange (array-like, optional):
            A two-element array with the (geometric) centers of the
            first and last pixel in the output vector. If not
            provided, assumed to be the same as the input range.
        newBorders (array-like, optional):
            An array with the borders of each pixel in the resampled
            vectors.
        newpix (:obj:`int`, optional): 
            Number of pixels for the output vector.  If not provided,
            assumed to be the same as the input vector.
        newLog (:obj:`bool`, optional):
            The output vector should be logarithmically binned.
        newdx (:obj:`float`, optional):
            The sampling step for the output vector. If
            `newLog=True`, this must be the change in the *logarithm*
            of :math:`x` for the output vector! If not provided, the
            sampling is set by the output range (see ``newRange``
            above) and number of pixels (see ``newpix`` above).
        base (:obj:`float`, optional):
            The base of the logarithm used for both input and output
            sampling, if specified. The default is 10; use
            ``numpy.exp(1)`` for natural logarithm.
        ext_value (:obj:`float`, optional):
            Set extrapolated values to the provided float. If set to
            None, values are just set to the linear extrapolation of
            the data beyond the provided limits; use `ext_value=None`
            with caution!
        conserve (:obj:`bool`, optional):
            Conserve the integral of the input vector.  For example, if
            the input vector is a spectrum in flux units, you should
            conserve the flux in the resampling; if the spectrum is in
            units of flux density, you do not want to conserve the
            integral.
        step (:obj:`bool`, optional):
            Treat the input function as a step function during the
            resampling integration.  If False, use a linear
            interpolation between pixel samples.
        covar (:obj:`bool`, optional):
            Calculate the covariance matrix between pixels in the
            resampled vector. Can only be used if ``step=True``. If
            no error vector is provided (``e``), the result is a
            *correlation* matrix.
    
    Attributes:
        x (`numpy.ndarray`_):
            The coordinates of the function on input.
        xborders (`numpy.ndarray`_):
            The borders of the input pixel samples.
        y (`numpy.ndarray`_):
            The function to resample.
        e (`numpy.ndarray`_):
            The 1-sigma errors in the function to resample.
        m (`numpy.ndarray`_):
            The boolean mask for the input function.
        outx (`numpy.ndarray`_):
            The coordinates of the function on output.
        outborders (`numpy.ndarray`_):
            The borders of the output pixel samples.
        outy (`numpy.ndarray`_):
            The resampled function.
        oute (`numpy.ndarray`_):
            The resampled 1-sigma errors.
        outf (`numpy.ndarray`_):
            The fraction of each output pixel that includes valid data
            from the input function.
        covar (:class:`~mangadap.util.covariance.Covariance`):
            The covariance or correlation matrices for the resampled
            vectors.

    Raises:
        ValueError:
            Raised if more the one of ``x``, ``xRange``, or
            ``xBorders`` are provided, if more the one of ``newx``,
            ``newRange``, or ``newBorders`` are provided, if ``y`` is
            a `numpy.ndarray`_, if ``y`` is not 1D or 2D, if the
            covariance is requested but ``step`` is False, if the
            shapes of the provided errors or mask do not match ``y``,
            if there is insufficient information to construct the
            input or output grid, or if either ``xRange`` or
            ``newRange`` are not two-element arrays.
    """
    def __init__(self, y, e=None, mask=None, x=None, xRange=None, xBorders=None, inLog=False,
                 newx=None, newRange=None, newBorders=None, newpix=None, newLog=True, newdx=None,
                 base=10.0, ext_value=0.0, conserve=False, step=True, covar=False):

        # Check operation can be performed and is not ill-posed
        if numpy.sum([inp is not None for inp in [x, xRange, xBorders]]) != 1:
            raise ValueError('One and only one of the x, xRange, and xBorders arguments should be '
                             'provided.')
        if numpy.sum([inp is not None for inp in [newx, newRange, newBorders]]) != 1:
            raise ValueError('One and only one of the newx, newRange, and newBorders arguments '
                             'should be provided.')
        if not isinstance(y, numpy.ndarray):
            raise ValueError('Input vector must be a numpy.ndarray!')
        if y.ndim > 2:
            raise ValueError('Input must be a 1D or 2D array!')
        if covar and not step:
            raise ValueError('Covariance is currently only calculated for step resampling.')

        # Setup the data, errors, and mask
        self.y = y.filled(0.0) if isinstance(y, numpy.ma.MaskedArray) else y.copy()
        self.twod = self.y.ndim == 2
        self.e = None if e is None \
                    else e.filled(0.0) if isinstance(e, numpy.ma.MaskedArray) else e.copy()
        self.m = numpy.zeros(self.y.shape, dtype=bool) if mask is None else mask

        # Check the shapes
        if self.e is not None and self.e.shape != self.y.shape:
            raise ValueError('Error array shape mismatched!')
        if self.m.shape != self.y.shape:
            raise ValueError('Mask array shape mismatched!')

        # Get the union of all the relevant masks
        if isinstance(y, numpy.ma.MaskedArray):
            self.m |= y.mask
        if e is not None and isinstance(e, numpy.ma.MaskedArray):
            self.m |= e.mask

        # Get the input coordinates
        nx = self.y.shape[-1] if x is None and xBorders is None else None
        self.x, self.xborders = self._coordinate_grid(x=x, rng=xRange, nx=nx, borders=xBorders,
                                                      log=inLog, base=base)

        # If conserving integral, assume input is integrated over pixel
        # width and convert to a density function (divide by pixel size)
        if conserve:
            self.y /= (numpy.diff(self.xborders)[None,:] if self.twod \
                                else numpy.diff(self.xborders))

        # Get the output coordinates
        nx = self.x.size \
                if newx is None and newBorders is None and newpix is None and newdx is None \
                else newpix
        self.outx, self.outborders = self._coordinate_grid(x=newx, rng=newRange, nx=nx,
                                                           borders=newBorders, dx=newdx,
                                                           log=newLog, base=base)

        if covar:
            A = self._resample_step_matrix()
            self.outy = numpy.dot(A, self.y.T).T
            self.outf = numpy.dot(A, numpy.logical_not(self.m.T).astype(int)).T \
                            / numpy.diff(self.outborders)[...,:]
            if self.e is None:
                # TODO: Repeat it N times if there are N y vectors?
                self.covar = Covariance.from_matrix_multiplication(A, numpy.ones_like(self.x)
                                                ).apply_new_variance(numpy.ones_like(self.outx))
                self.oute = None
            else:
                if self.twod:
                    self.covar = numpy.empty(self.y.shape[0], dtype=object)
                    for i in range(self.y.shape[0]):
                        self.covar[i] = Covariance.from_matrix_multiplication(A,
                                            numpy.square(self.e[i])).full()
                    self.covar = Covariance(self.covar, impose_triu=True)
                    self.oute = numpy.sqrt(self.covar.variance().T)
                else:
                    self.covar = Covariance.from_matrix_multiplication(A, numpy.square(self.e))
                    self.oute = numpy.sqrt(self.covar.variance())
        else:
            # Perform the resampling
            self.covar = None
            self.outy = self._resample_step(self.y) if step else self._resample_linear(self.y)
       
            # The mask and errors are always interpolated as a step function
            self.oute = None if self.e is None else self._resample_step(self.e, quad=True)
        
            self.outf = self._resample_step(numpy.invert(self.m).astype(int)) \
                            / numpy.diff(self.outborders)

        # Do not conserve the integral over the size of the pixel
        if not conserve:
            self.outy /= (numpy.diff(self.outborders)[None,:] if self.twod \
                            else numpy.diff(self.outborders))
            if self.oute is not None:
                self.oute /= (numpy.diff(self.outborders)[None,:] if self.twod \
                                    else numpy.diff(self.outborders))
                if self.covar is not None:
                    self.covar = self.covar.apply_new_variance(numpy.square(self.oute.T))

        # Set values for extrapolated regions
        if ext_value is not None:
            indx = (self.outborders[:-1] < self.xborders[0]) \
                        | (self.outborders[1:] > self.xborders[-1]) 
            if numpy.sum(indx) > 0:
                self.outy[...,indx] = ext_value
                self.outf[...,indx] = 0.
                if self.oute is not None:
                    self.oute[...,indx] = 0.

    @staticmethod
    def _coordinate_grid(x=None, rng=None, nx=None, dx=None, borders=None, log=False, base=10.0):
        """
        Use the provided information to construct the coordinate grid
        and the grid borders.
        """
        if x is not None and borders is not None:
            raise ValueError('Both x and borders provided.  Do not need to call _coordinate_grid, '
                             'but also _coordinate_grid does not check that x and borders are '
                             'consistenet with one another.')
        if (x is not None or borders is not None) and rng is not None:
            warnings.warn('Provided both x or borders and the range.  Ignoring range.')
        if x is None and borders is not None:
            # Use the borders to set the centers
            return borders_to_centers(borders, log=log), borders
        if x is not None and borders is None:
            # Use the centers to set the borders
            return x, centers_to_borders(x, log=log)

        # After this point, both x and borders should be None
        assert x is None and borders is None, 'Coding logic error'

        if rng is None and nx is None:
            raise ValueError('Insufficient input to construct coordinate grid.')

        if rng is None:
            # Just set the result to a uniform pixel grid
            return numpy.arange(nx, dtype=float) + 0.5, numpy.arange(nx+1, dtype=float)

        # After this point, rng cannot be None
        assert rng is not None, 'Coding logic error'

        if dx is not None and nx is not None:
            warnings.warn('Provided rng, dx, and nx, which over-specifies the grid; rng and nx '
                          'take precedence.')
        if nx is not None:
            borders = grid_borders(rng, nx, log=log, base=base)[0]
            return borders_to_centers(borders, log=log), borders

        nx, _rng = grid_npix(rng=rng, dx=dx, log=log, base=base)
        borders = grid_borders(_rng, nx, log=log, base=base)[0]
        return borders_to_centers(borders, log=log), borders

    def _resample_linear(self, v, quad=False):
        """Resample the vectors."""

        # Combine the input coordinates and the output borders
        combinedX = numpy.append(self.outborders, self.x)
        srt = numpy.argsort(combinedX)
        combinedX = combinedX[srt]

        # Get the indices where the data should be reduced
        border = numpy.ones(combinedX.size, dtype=bool)
        border[self.outborders.size:] = False
        k = numpy.arange(combinedX.size)[border[srt]]

        # Calculate the integrand
        if self.twod:
            # Linearly interpolate the input function at the output border positions
            interp = interpolate.interp1d(self.x, v, axis=-1, assume_sorted=True,
                                          fill_value='extrapolate')
            combinedY = numpy.append(interp(self.outborders), v, axis=-1)[:,srt]
            integrand = (combinedY[:,1:]+combinedY[:,:-1])*numpy.diff(combinedX)[None,:]/2.0
        else:
            # Linearly interpolate the input function at the output border positions
            interp = interpolate.interp1d(self.x, v, assume_sorted=True,
                                          fill_value='extrapolate')
            combinedY = numpy.append(interp(self.outborders), v)[srt]
            integrand = (combinedY[1:]+combinedY[:-1])*numpy.diff(combinedX)/2.0

        if quad:
            integrand = numpy.square(integrand)

        # Use reduceat to calculate the integral
        out = numpy.add.reduceat(integrand, k[:-1], axis=-1) if k[-1] == combinedX.size-1 \
                        else numpy.add.reduceat(integrand, k, axis=-1)[...,:-1]
    
        return numpy.sqrt(out) if quad else out

    def _resample_step(self, v, quad=False):
        """Resample the vectors."""

        # Convert y to a step function
        #  - repeat each element of the input vector twice
        _v = numpy.repeat(v, 2, axis=1) if self.twod else numpy.repeat(v, 2)
        #  - repeat each element of the border array twice, and remove
        #  the first and last elements
        _x = numpy.repeat(self.xborders, 2)[1:-1]

        # Combine the input coordinates and the output borders into a
        # single vector
        indx = numpy.searchsorted(_x, self.outborders)
        combinedX = numpy.insert(_x, indx, self.outborders)

        # Insert points at the borders of the output function
        v_indx = indx.copy()
        v_indx[indx >= _v.shape[-1]] = -1
        combinedY = numpy.array([ numpy.insert(__v, indx, __v[v_indx]) for __v in _v ]) \
                            if self.twod else numpy.insert(_v, indx, _v[v_indx])

        # Calculate the integrand
        integrand = combinedY[:,1:]*numpy.diff(combinedX)[None,:] if self.twod else \
                        combinedY[1:]*numpy.diff(combinedX)
        if quad:
            integrand = numpy.square(integrand)

        # Get the indices where the data should be reduced
        border = numpy.insert(numpy.zeros(_x.size, dtype=bool), indx,
                              numpy.ones(self.outborders.size, dtype=bool))
        k = numpy.arange(combinedX.size)[border]

        # Use reduceat to calculate the integral
        out = numpy.add.reduceat(integrand, k[:-1], axis=-1) if k[-1] == combinedX.size-1 \
                    else numpy.add.reduceat(integrand, k, axis=-1)[...,:-1]
        return numpy.sqrt(out) if quad else out

    def _resample_step_matrix(self):
        r"""
        Build a matrix such that

        .. math::
            y = \mathbf{A} x

        where :math:`x` is the input vector, :math:`y` is the resampled
        vector, and :math:`\mathbf{A}` is the matrix operations that
        resamples :math:`x`.
        """
        ny = self.outx.size
        nx = self.x.size
        dx = numpy.diff(self.xborders)

        # Repeat each element of the border array twice, and remove the
        # first and last elements
        _p = numpy.repeat(numpy.arange(self.x.size), 2)
        _x = numpy.repeat(self.xborders, 2)[1:-1]

        # Combine the input coordinates and the output borders into a
        # single vector
        indx = numpy.searchsorted(_x, self.outborders)
        combinedX = numpy.insert(_x, indx, self.outborders)

        # Insert points at the borders of the output function
        p_indx = indx.copy()
        p_indx[indx >= _p.shape[-1]] = -1
        combinedP = numpy.insert(_p, indx, _p[p_indx])

        # Get the indices where the data should be reduced
        border = numpy.insert(numpy.zeros(_x.size, dtype=bool), indx,
                              numpy.ones(self.outborders.size, dtype=bool))
        nn = numpy.where(numpy.invert(border))[0][::2]
        k = numpy.zeros(len(combinedX), dtype=int)
        k[border] = numpy.arange(numpy.sum(border))
        k[nn-1] = k[nn-2]
        k[nn] = k[nn-1]
        start,end = numpy.where(border)[0][[0,-1]]

        # Calculate the fraction of each pixel into each output pixel
        fraction = numpy.diff(combinedX[start:end+1])
        # Construct the output matrix
        indx = fraction > 0
        A = numpy.zeros((ny, nx), dtype=float)
        A[k[start:end][indx], combinedP[start:end][indx]] = fraction[indx]
        return A






#!/usr/bin/env python3

import numpy
from os import remove
import os.path
from time import clock

#-----------------------------------------------------------------------------

class BasePar:
    def __init__(self, inp):
        print('running base init')
        print('')
        self.pars = inp
        self.write()
    def write(self):
        print('running base write')
        print(self.pars)
        print('')

class DerivedPar(BasePar):
    def __init__(self, item1, item2, item3):
        print('running derived init')
        print('')
        inp = [ item1, item2, item3 ]
        BasePar.__init__(self, inp)

def run_test():
    x = DerivedPar('test', 1.0, 0)
    x.write()
    print('')

    x.pars[0] = 'derivedtest'
    x.write()
    print('')

    y = BasePar(['basetest', 1.0, 0])
    y.write()
    print('')

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = clock()
    run_test()
    print('Elapsed time: {0} seconds'.format(clock() - t))






from IPython import embed

import numpy

from mangadap.contrib import xjmc


def test_component_select():
    comp = numpy.array([ 0,  1,  2,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 14,
                        15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26,  8])

    valid = numpy.ones(comp.size, dtype=bool)

    _comp, comp_map = xjmc._reset_components(comp, valid)
    assert numpy.array_equal(comp, _comp), 'Component list should not have changed.'

    # Remove one template that doesn't cause the removal of an entire
    # component.
    valid[3] = False
    _comp, comp_map = xjmc._reset_components(comp, valid)
    assert numpy.array_equal(comp[valid], _comp), 'Component IDs should not have changed'
    valid[3] = True

    # Remove a template that constitutes an entire component
    valid[12] = False
    _comp, comp_map = xjmc._reset_components(comp, valid)
    assert len(_comp) == len(comp)-1, 'Component was not removed'
    assert numpy.array_equal(_comp[12:-1]+1, comp[13:-1]), \
            'Components were not reordered correctly.'
    valid[12] = True


def test_constraint_select():
    comp = numpy.array([ 0, 1, 2, 2, 3, 4])
    valid = numpy.ones(comp.size, dtype=bool)

    A_ineq = numpy.array([[ 0., 0.91, 0., 0., 0., 0., 0., -1., 0., 0.],
                          [ 0., -1.1, 0., 0., 0., 0., 0.,  1., 0., 0.]])
    b_ineq = numpy.zeros(2, dtype=float)

    # Removes a template that has other templates in this component, so the
    # return arrays should be identical.
    valid[2] = False
    _A, _b = xjmc._reset_kinematic_constraints(comp, valid, A_ineq, b_ineq)
    assert numpy.array_equal(A_ineq, _A) and numpy.array_equal(b_ineq, _b), \
            'No components should have been removed'
    valid[2] = True

    # Removes the only template in a component, but the component is not
    # constrained by A,b. So the shape of the return A matrix should have the
    # relevant component columns removed.
    valid[1] = False
    _A, _b = xjmc._reset_kinematic_constraints(comp, valid, A_ineq, b_ineq)
    assert numpy.array_equal(A_ineq[:,numpy.repeat([True, False, True, True, True],2)], _A) \
                and numpy.array_equal(b_ineq, _b), 'No components should have been removed'
    valid[1] = True

    # Removes a template constituting an entire component constrained by A,b,
    # such that no constraints remain
    valid[0] = False
    _A, _b = xjmc._reset_kinematic_constraints(comp, valid, A_ineq, b_ineq)
    assert _A is None and _b is None, 'No constraints should remain'
    valid[0] = True

    # Try with more constraints    
    A_ineq = numpy.array([[ 0., 0.91, 0., 0., 0.,   0., 0., -1., 0., 0.],
                          [ 0., -1.1, 0., 0., 0.,   0., 0.,  1., 0., 0.],
                          [ 0.,   0., 0., 0., 0., 0.91, 0., -1., 0., 0.],
                          [ 0.,   0., 0., 0., 0., -1.1, 0.,  1., 0., 0.]])
    b_ineq = numpy.zeros(4, dtype=float)

    # Removes the only template in a component, but the component is not
    # constrained by A,b. So the shape of the return A matrix should have the
    # relevant component columns removed.
    valid[1] = False
    _A, _b = xjmc._reset_kinematic_constraints(comp, valid, A_ineq, b_ineq)
    assert numpy.array_equal(A_ineq[:,numpy.repeat([True, False, True, True, True],2)], _A) \
                and numpy.array_equal(b_ineq, _b), 'No components should have been removed'
    valid[1] = True

    # Removes a template constituting an entire component constrained by 2 of
    # the 4 constraints defined by A,b, such only 2 remain in the output.
    valid[0] = False
    _A, _b = xjmc._reset_kinematic_constraints(comp, valid, A_ineq, b_ineq)
    assert numpy.array_equal(A_ineq[2:,numpy.repeat([False, True, True, True, True],2)], _A) \
                and numpy.array_equal(b_ineq[2:], _b), 'No components should have been removed'
    valid[0] = True

    # Removes a template constituting an entire component that included in all
    # the constraints defined by A,b, such that no constraints remain
    valid[4] = False
    _A, _b = xjmc._reset_kinematic_constraints(comp, valid, A_ineq, b_ineq)
    assert _A is None and _b is None, 'No constraints should remain'
    valid[4] = True





import time
import numpy

from matplotlib import pyplot

from astropy.io import fits

from mangadap.config import manga, defaults
from mangadap.datacube import MaNGADataCube
from mangadap.util.sampling import Resample, grid_borders
from mangadap.proc.bandpassfilter import passband_integral

try:
    import spectres
except:
    spectres = None

#-----------------------------------------------------------------------------

def get_redshift(plt, ifu, drpall_file=None):
    """
    Get the redshift of a galaxy from the DRPall file.

    Args:
        plt (:obj:`int`):
            Plate number
        ifu (:obj:`int`):
            IFU identifier
        drapall_file (:obj:`str`, optional):
            DRPall file. If None, attempts to use the default path to
            the file using environmental variables.
    
    Returns:
        :obj:`float`: The redshift to the galaxy observed by the
        provided PLATEIFU.
    """
    if drpall_file is None:
        drpall_file = manga.drpall_file()
    if not drpall_file.exists():
        raise FileNotFoundError(f'Could not find DRPall file: {drpall_file}')
    hdu = fits.open(str(drpall_file))
    indx = hdu[1].data['PLATEIFU'] == '{0}-{1}'.format(plt, ifu)
    return hdu[1].data['NSA_Z'][indx][0]

#-----------------------------------------------------------------------------

def resample_test():

    # Read the example datacube and get the expected redshift You can download
    # these data using
    # https://github.com/sdss/mangadap/blob/master/download_test_data.py
    plt = 7815
    ifu = 3702
    drpver = 'v3_1_1'
    directory_path = defaults.dap_source_dir() / 'data' / 'remote'
    cube = MaNGADataCube.from_plateifu(plt, ifu, directory_path=directory_path)
    drpall_file = directory_path / f'drpall-{drpver}.fits'
    z = get_redshift(plt, ifu, drpall_file)

    # Pull out two example spectra from the datacube
    old_wave = cube.wave
    old_flux = numpy.ma.MaskedArray(cube.flux[10,10:12,:], mask=cube.mask[10,10:12,:] > 0)
    old_flux[:,(old_wave > 5570) & (old_wave < 5586)] = numpy.ma.masked
    old_ferr = numpy.ma.power(cube.ivar[10,10:12,:], -0.5)

    if spectres is not None:
        # Use spectres to resample the spectrum, ignoring last pixel
        indx = (old_wave > old_wave[0]/(1+z)) & (old_wave < old_wave[-2]/(1+z))
        t = time.perf_counter()
        new_flux_spectres = numpy.empty((old_flux.shape[0], numpy.sum(indx)), dtype=float)
        new_ferr_spectres = numpy.empty((old_flux.shape[0], numpy.sum(indx)), dtype=float)
        for i in range(old_flux.shape[0]):
            new_flux_spectres[i,:], new_ferr_spectres[i,:] \
                    = spectres.spectres(old_wave[indx], old_wave/(1+z), old_flux[i,:].filled(0.0),
                                        spec_errs=old_ferr[i,:].filled(0.0))
        print('SpectRes Time: ', time.perf_counter()-t)

    # Use a brute-force integration of the spectra to resample the spectrum
    t = time.perf_counter()
    borders = grid_borders(numpy.array([old_wave[0],old_wave[-1]]), old_wave.size, log=True)[0]
    _p = numpy.repeat(borders, 2)[1:-1].reshape(-1,2)
    new_flux_brute = numpy.array([passband_integral(old_wave/(1+z), f, passband=_p, log=True)
                                        for f in old_flux.filled(0.0)])
    new_flux_brute /= (_p[:,1]-_p[:,0])[None,:]
    print('Brute Force Time: ', time.perf_counter()-t)

    # Use the Resample class to resample the spectrum
    t = time.perf_counter()
    r = Resample(old_flux, e=old_ferr, x=old_wave/(1+z), newRange=[old_wave[0], old_wave[-1]],
                 inLog=True, newLog=True)
    print('Resample Time: ', time.perf_counter()-t)

    # Estimate the differences between the resampling methods (these should all
    # be the same to nearly numerical accuracy)
    print('Mean diff:')
    if spectres is not None:
        print('    spectres - brute    = {0:.5e}'.format(
                numpy.mean(numpy.absolute(new_flux_spectres-new_flux_brute[:,indx]))))
        print('    spectres - resample = {0:.5e}'.format(
                numpy.mean(numpy.absolute(new_flux_spectres-r.outy[:,indx]))))
    print('    brute - resample    = {0:.5e}'.format(
            numpy.mean(numpy.absolute(new_flux_brute-r.outy))))

    # Plot the original and resampled versions for all spectra.  The resampled
    # versions should all be indistinguishable.
    for i in range(old_flux.shape[0]):
        pyplot.plot(old_wave/(1+z), old_flux[i,:], label='Data')
        if spectres is not None:
            pyplot.plot(old_wave[indx], new_flux_spectres[i,:], label='spectres')
        pyplot.plot(old_wave, new_flux_brute[i,:], label='brute')
        pyplot.plot(r.outx, r.outy[i,:], label='Resample')
        pyplot.plot(r.outx, r.outf[i,:], label='Good-pixel Mask')
        pyplot.legend()
        pyplot.xlabel('Wavelength')
        pyplot.ylabel('Flux')
        pyplot.show()

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    resample_test()






import os
import time
import warnings
import argparse

import numpy

try:
    import pbs.queue
except:
    warnings.warn('Could not import pbs.queue!  Any cluster submission will fail!', ImportWarning)

def parse_args():

    # Declare the ArgumentParser object
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--output_path', type=str, help='path for output', default=os.getcwd())
    parser.add_argument('--output_root', type=str, help='root name for outputfiles', default=None)

    parser.add_argument('--label', type=str, help='label for cluster job', default='daptest')
    parser.add_argument('--nodes', type=int, help='number of nodes to use in cluster', default=1)
    parser.add_argument('--cpus', type=int,
                        help='number of cpus to use per node.  Default is to use all available'
                             '; otherwise, set to minimum of provided number and number of '
                             'processors per node', default=None)
    parser.add_argument('--fast', dest='qos', type=str, help='qos state', default=None)
    parser.add_argument('--umask', type=str, help='umask bit for cluster job', default='0027')
    parser.add_argument('--walltime', type=str, help='walltime for cluster job',
                        default='240:00:00')
    parser.add_argument('--toughness', dest='hard', action='store_false', default=True,
                        help='turn off hard keyword for cluster submission')

    parser.add_argument('--create', action='store_true', default=False,
                        help='use the pbs package to create the cluster scripts')
    parser.add_argument('--submit', action='store_true', default=False,
                        help='submit the scripts to the cluster')
    
    return parser.parse_args()


#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
def main():

    # Read the arguments
    args = parse_args()
    if args.submit and not args.create:
        args.create = True

    # Hard-wired cases to run
    sc_tpl = ['MILESHC']*16
    sc_vsr = [2] + [4]*15
    sc_deg = [8]*16

    el_tpl = [None, 'MILESHC'] + ['MASTARHC2']*14
    el_vsr = [None, 4] + [1]*14
    el_deg = [None, 8] + (numpy.arange(14)+1).tolist()
    el_band = [None] + ['ELBMPL9']*15
    el_list = [None] + ['ELPMPL9']*15

    # Hard-wired location of script
    fit_script = os.path.abspath('fit_selected_spectra.py')
    # Hard-wired files to use
    benchmark = os.path.abspath('benchmark_spectra_v2.fits')
    flags = os.path.abspath('representative_spectra_flags_v2_fitall.db')

    # Output path for results
    output_path = os.path.abspath(args.output_path)
    if not os.path.isdir(output_path):
        os.makedirs(output_path)

    # Path for script and log files
    log_path = os.path.join(output_path, 'log')
    if not os.path.isdir(log_path):
        os.makedirs(log_path)

    # Build the queue if requested
    if args.create:
        queue = pbs.queue(verbose=True)
        ppn = 16
        cpus = ppn if args.cpus is None else min(args.cpus, ppn)
        queue.create(label=args.label, nodes=args.nodes, qos=args.qos, umask=args.umask,
                     walltime=args.walltime, ppn=ppn, cpus=cpus)

    nfits = len(sc_tpl)

    for i in range(nfits):
        ofile_root = '{0}_{1}_{2}_{3}'.format(args.output_root, sc_tpl[i], sc_vsr[i], sc_deg[i])
        if el_tpl[i] is not None:
            ofile_root += '_{0}_{1}_{2}_{3}_{4}'.format(el_tpl[i], el_vsr[i], el_deg[i],
                                                        el_band[i], el_list[i])

        # Set the names for the script, stdout, and stderr files
        scriptfile = os.path.join(log_path, ofile_root)
        stdoutfile = '{0}.out'.format(scriptfile)
        stderrfile = '{0}.err'.format(scriptfile)

        # Write the file
        file = open(scriptfile, 'w')
        file.write('# Auto-generated batch file\n')
        file.write('# {0}\n'.format(time.strftime("%a %d %b %Y %H:%M:%S",time.localtime())))
        file.write('\n')

        file.write('touch {0}\n'.format('{0}.started'.format(scriptfile)))
        file.write('\n')

        command = 'OMP_NUM_THREADS=1 python3 {0}'.format(fit_script)
        command += ' {0} {1}.fits.gz'.format(benchmark, ofile_root)
        command += ' --output_root {0}'.format(output_path)
        command += ' --spec_flags {0}'.format(flags)
        command += ' --sc_tpl {0}'.format(sc_tpl[i])
        command += ' --sc_vsr {0}'.format(sc_vsr[i])
        command += ' --sc_deg {0}'.format(sc_deg[i])
        if el_tpl[i] is None:
            command += ' --sc_only'
        else:
            command += ' --el_tpl {0}'.format(el_tpl[i])
            command += ' --el_vsr {0}'.format(el_vsr[i])
            command += ' --el_deg {0}'.format(el_deg[i])
            command += ' --el_band {0}'.format(el_band[i])
            command += ' --el_list {0}'.format(el_list[i])
        file.write('{0}\n'.format(command))
        file.write('\n')

        file.write('touch {0}\n'.format('{0}.done'.format(scriptfile)))
        file.write('\n')
        file.close()
        
        if args.create:
            queue.append('source {0}'.format(scriptfile), outfile=stdoutfile, errfile=stderrfile)

    # Submit the queue to the cluster
    if args.create:
        queue.commit(hard=args.hard, submit=args.submit)

if __name__ == '__main__':
    main()


"""
Base class for a datacube

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

# TODO: Pilfer the pypeit.DataContainer for this.

# TODO: Force the datacube data arrays to be read-only?

import warnings

from IPython import embed

import numpy

from scipy import sparse, interpolate

from astropy.io import fits
import astropy.constants

from ..util.bitmask import BitMask
from ..util.mapping import permute_wcs_axes
from ..util.covariance import Covariance
from ..util.pixelmask import SpectralPixelMask
from ..util.sampling import angstroms_per_pixel
from ..util.geometry import polygon_area

class DataCube:
    r"""
    Base container class for a rectilinear datacube.

    Datacubes have three axes: two spatial coordinates and one spectral. The
    datacubes are expected to be rectilinear; i.e., each spatial position has
    the same spectral range and each wavelength channel has the same spatial
    coordinates.

    On input, the ordering of the three dimensions can be arbitrary; however,
    the axes are re-ordered for the internal attributes such that wavelengths
    are ordered along the last axis, with the first two axes being the
    spatial coordinates. Nominally the spatial coordinates are ordered
    predominantly coincident with right-ascension (E toward smaller pixels
    numbers in axis 0) and declination (N toward larger pixel numbers in axis
    1); however, rotation of the datacube spatial coordinates can be non-zero
    with respect to the celestial coordinates, as given by the provided
    world-coordinate system.

    The wavelength vector applicable to all spatial positions can either be
    provided directly or constructed from the provided WCS. Any directly
    provided ``wave`` vector takes precedence over the WCS.

    Spatial covariance/correlation can be provided via the ``covar`` keyword
    argument. If provided as a single array, the provided array is used to
    define the *correlation* matrix and is assumed to be identical for all
    wavelength channels. More options are available if the input is provided
    as a :class:`~mangadap.util.covariance.Covariance` object. The ordering
    of the covariance/correlation matrix is expected to provide the
    correlation between the row-major flattened version of the spatial
    coordinates. That is, for a datacube with spatial shape
    :math:`(N_x,N_y)`, the covariance/correlation value at location
    :math:`i,j` is the correlation between pixels at 2D locations
    :math:`(i_x,i_y)` and :math:`(j_x,j_y)`, where

    .. math::

        \begin{array}{rcl}
        i_x & = & \lfloor i/N_y \rfloor, \\
        i_y & = & i - i_x N_y, \\
        j_x & = & \lfloor j/N_y \rfloor, {\rm and} \\
        j_y & = & j - j_x N_y.
        \end{array}

    If the values of ``axes`` indicates that the provided flux array should
    have it's spatial axes transposed (i.e. ``axes[0] > axes[1]``), the
    provided covariance/correlation data is appropriately restructured to
    correspond to the wavelength channels in :attr:`flux` (again assuming a
    flattened row-major memory block); see
    :func:`~mangadap.util.covariance.Covariance.transpose_raw_shape`.
    
    See :ref:`datacube_subclass` for instructions on subclassing this object for
    non-MaNGA datacubes.

    Args:
        flux (`numpy.ndarray`_):
            Data array with the flux as a function of spatial and
            spectral position. Shape should be :math:`(N_x, N_y,
            N_\lambda)`.
        wave (`numpy.ndarray`_, optional):
            The wavelength vector associated with each spatial
            position. Shape is :math:`(N_\lambda,)`. If None, ``wcs``
            must be provided such that the wavelength vector can be
            constructed.
        ivar (`numpy.ndarray`_, optional):
            The inverse variance of the flux array. Shape must be the
            same as ``flux``.  If None, errors are ignored.
        mask (`numpy.ndarray`_, optional):
            Mask array for the flux measurements. Can be a boolean
            mask (False=unmasked,good; True=masked,bad) or an integer
            array associated with the provided
            :class:`mangadap.util.bitmask.BitMask` object (see
            ``bitmask``). Shape must be the same as ``flux``.
        bitmask (:class:`mangadap.util.bitmask.BitMask`, optional):
            Object used to select and toggle masked pixels from
            :attr:`mask`. If None, any provided ``mask`` must be a
            boolean array or can be successfully converted to one.
        sres (`numpy.ndarray`_, optional):
            The spectral resolution, :math:`R = \lambda /
            \Delta\lambda`. Can be a single vector with shape
            :math:`(N_\lambda,)`, if the spectral resolution is
            independent of spatial position, or a 3D datacube with a
            spatially and spectrally dependent resolution with a
            shape that matches ``flux``. If None, spectral resolution
            is ignored.
        covar (array-like, :class:`mangadap.util.covariance.Covariance`, optional):
            The spatial covariance in one or more wavelength
            channels. If None, no covariance is assumed. See usage
            and interpretation in class description.
        axes (:obj:`list`, optional):
            The axes with the :math:`x`, :math:`y`, and
            :math:`\lambda` in the provided flux array. For example,
            ``[2,1,0]`` means the spectra are organized along the
            first axis.
        wcs (`astropy.wcs.WCS`_, optional):
            World-coordinate system for the 3D datacube. If None,
            spatial coordinates are set to be centered with a
            pixelscale in arcsec (see ``pixelscale``).
        pixelscale (:obj:`int`, :obj:`float`, optional):
            Spatial extent in arcsec of each spaxel in the datacube,
            assumed to be square. Superseded by ``wcs``, if the
            latter is provided. If the ``wcs`` is not provided and
            the pixelscale is not provided upon instantiation, it is
            set to unity.
        log (:obj:`bool`, optional):
            Flag that the datacube spectral pixels are binned
            logarithmically in wavelength.
        meta (:obj:`dict`, optional):
            A free-form dictionary used to hold metadata relevant to the
            datacube. Metadata required by analysis modules are indicated where
            relevant. If None, :attr:`meta` is instantiated as an empty
            dictionary and populated with the bare minimum needed to execute the
            DAP.  NOTE: Currently this is a required argument because an
            estimate of the redshift is needed to execute the DAP.
        prihdr (`astropy.io.fits.Header`_, optional):
            Primary header read from datacube fits file. If None,
            instantiated as an empty `astropy.io.fits.Header`_.
        fluxhdr (`astropy.io.fits.Header`_, optional):
            Header specifically for the flux extension of the
            datacube fits file. If None, set to be a copy of the
            primary header.

    Raises:
        ValueError:
            Raised if the wavelength vector cannot be produced, if
            the WCS has the wrong dimensionality, or if there are any
            shape mismatches between the input data arrays.
        TypeError:
            Raised if the input metadata are not provided as a
            dictionary.

    Attributes:
        original_axes (`numpy.ndarray`_):
            Original axis order.
        shape (:obj:`tuple`):
            Datacube shape.
        spatial_shape (:obj:`tuple`):
            Shape of the datacube spatial axes .
        nwave (:obj:`int`):
            Number of wavelength channels.
        spatial_index (`numpy.ndarray`_):
            Array of tuples with the spatial indices of each spectrum
            in the flattened datacube.
        wcs (`astropy.wcs.WCS`_):
            Datacube world-coordinate system.  Can be None.
        pixelscale (:obj:`float`):
            Spatial extent in arcsec of each spaxel in the datacube,
            assumed to be square. Only used if :attr:`wcs` is None.
        bitmask (:class:`mangadap.util.bitmask.BitMask`):
            Object used to select and toggle masked pixels from
            :attr:`mask`.  Can be None.
        log (:obj:`bool`):
            Flag that the datacube spectral pixels are binned
            logarithmically in wavelength.
        meta (:obj:`dict`):
            A free-form dictionary used to hold meta data relevant to
            the datacube. Metadata required by analysis modules are
            indicated where relevant. If no metadata has been
            defined, :attr:`meta` is instantiated as an empty
            dictionary.
        wave (`numpy.ndarray`_):
            Wavelength vector applicable to all spatial positions.
        flux (`numpy.ndarray`_):
            Datacube flux array.
        ivar (`numpy.ndarray`_):
            Datacube flux inverse variance.  Can be None.
        mask (`numpy.ndarray`_):
            Datacube mask.  Can be None.
        sres (`numpy.ndarray`_):
            Datacube spectral resolution.  Can be None.
        covar (:class:`mangadap.util.covariance.Covariance`):
            Datacube spatial covariance.  Can be None.
        prihdr (`astropy.io.fits.Header`_):
            Primary header for the datacube. If not provided on
            instantiation, set to an empty `astropy.io.fits.Header`_.
        fluxhdr (`astropy.io.fits.Header`_):
            Header specifically for the flux array. If not provided
            on instantiation, set to be a copy of :attr:`prihdr`.
        rss (:class:`mangadap.spectra.rowstackedspectra.RowStackedSpectra`):
            The source row-stacked spectra used to build the
            datacube.
        sigma_rho (:obj:`float`):
            The :math:`\sigma_{\rho}` of the Gaussian function used
            to approximate the trend of the correlation coefficient
            with spaxel separation. Used to construct the approximate
            correlation matrix (see
            :func:`approximate_correlation_matrix`).
        correl_rlim (:obj:`float`):
            The limiting radius of the image reconstruction
            (Gaussian) kernel in arcseconds. Used to construct the
            approximate correlation matrix (see
            :func:`approximate_correlation_matrix`).
        approx_correl (:class:`~mangadap.util.covariance.Covariance`):
            Approximate correlation matrix; see
            :func:`approximate_correlation_matrix`.
    """

    instrument = None
    """
    The name of the instrument used to collect the data.  This is *only* used in
    construction of the root name for output files.  See
    :func:`~mangadap.datacube.datacube.DataCube.output_root`.
    """

    # TODO: Add reconstructed PSF?
    def __init__(self, flux, wave=None, ivar=None, mask=None, bitmask=None, sres=None, covar=None,
                 axes=[0,1,2], wcs=None, pixelscale=None, log=True, meta=None, prihdr=None,
                 fluxhdr=None, name=None):

        # Set an identifier for the datacube
        self.name = name

        if wcs is None and wave is None:
            raise ValueError('Must either provide a single wavelength vector or a WCS that can '
                             'be used to construct it.')
        if wcs is not None and wcs.wcs.naxis != 3:
            raise ValueError('Provided WCS object must define the coordinate system for each of '
                             'the three datacube axes.')
        self.wcs = None if wcs is None else permute_wcs_axes(wcs, axes)
        self.pixelscale = self._get_pixelscale() if self.wcs is not None and pixelscale is None \
                                else (1 if pixelscale is None else float(pixelscale))

        self.original_axes = numpy.atleast_1d(axes).copy()
        # Re-order so that axes are x, y, lambda
        self.flux = flux.transpose(self.original_axes)
        self.shape = self.flux.shape
        self.spatial_shape = self.shape[:-1]
        self.nwave = self.shape[-1]

        # TODO: Not sure this is useful
        self.spatial_index = numpy.array([(i,j) for i,j in zip(*numpy.unravel_index(
                                                numpy.arange(self.nspec), self.spatial_shape))])

        self.meta = {} if meta is None else meta
        if not isinstance(self.meta, dict):
            raise TypeError('Metadata must be provided as a dictionary.')
        self.populate_metadata()

        self.wave = None if wave is None else numpy.atleast_1d(wave)
        if self.wave is None:
            self.wave = self._get_wavelength_vector()
        if self.wave.shape != (self.nwave,):
            raise ValueError('Wavelength vector shape is incorrect ')
        self.log = log
        # TODO: Check the above against the wavelength vector and/or
        # set the value automatically based on the input.
        
        self.ivar = None 
        if ivar is not None:
            self.ivar = ivar.transpose(self.original_axes)
            if self.ivar.shape != self.shape:
                raise ValueError('Inverse variance array has incorrect shape.')

        self.mask = None
        if mask is not None:
            self.mask = mask.transpose(self.original_axes)
            if self.mask.shape != self.shape:
                raise ValueError('Mask array has incorrect shape.')
        self.bitmask = None
        if bitmask is not None:
            if self.mask is None:
                warnings.warn('Bitmask provided but no mask data available.')
            if isinstance(bitmask, BitMask):
                self.bitmask = bitmask
            else:
                warnings.warn('Bitmasks must be BitMask objects; provided bitmask is ignored.')
        if self.bitmask is None:
            # Ensure mask is a boolean array
            self.mask = self.mask.astype(bool)

        self.redux_bitmask = None
        self.redux_qual_key = None
        self.redux_qual_flag = None
        
        self.sres = None
        if sres is not None:
            self.sres = numpy.tile(sres, self.spatial_shape+(1,)) \
                            if sres.ndim == 1 else sres.transpose(self.original_axes)
            if self.sres.shape != self.shape:
                raise ValueError('Spectral resolution array has incorrect shape.')

        self.covar = None
        if covar is not None:
            spatial_transpose = self.original_axes[0] > self.original_axes[1]
            raw_shape = self.spatial_shape[::-1] if spatial_transpose else self.spatial_shape
            self.covar = covar if isinstance(covar, Covariance) \
                            else Covariance.from_array(covar, raw_shape=raw_shape)
            if spatial_transpose:
                self.covar = self.covar.transpose_raw_shape()

        # Allocate attributes for primary and flux array fits headers
        self.prihdr = fits.Header() if prihdr is None else prihdr
        self.fluxhdr = self.prihdr.copy() if fluxhdr is None else fluxhdr

        # Allow for a RowStackedSpectrum counterpart
        self.rss = None

        # For the approximate covariance matrix calculations
        self.sigma_rho = None
        self.correl_rlim = None
        self.approx_correl = None

    @property
    def file_path(self):
        """
        Return the full path to the input file.

        .. warning::

            The required attributes for this method are *not* defined by the
            base class.  If the attributes ``directory_path`` or ``file_name``
            do not exist or if either of them are None, this returns None.
        """
        try:
            return self.directory_path / self.file_name
        except:
            warnings.warn('File path for input cube not defined!')
            return None

        # TODO: Unsure why, but the code below can lead to maximum recursion depth error!
#        if not hasattr(self, 'directory_path') or not hasattr(self, 'file_path') \
#                or self.directory_path is None or self.file_name is None:
#            return None
#        return self.directory_path / self.file_name

    @property
    def output_root(self):
        if self.instrument is None and self.name is None:
            raise ValueError('Output root is undefined.  Must provide instrument and/or name.')
        if self.instrument is not None and self.name is not None:
            return f'{self.instrument}-{self.name}'
        if self.instrument is not None:
            return f'{self.instrument}'
        return f'{self.name}'

    @property
    def can_analyze(self):
        """
        Confirm that the DAP can analyze the datacube.

        The only requirement is:

            - the :math:`cz` velocity (``'vel'`` in :attr:`meta`) must be
              greater than -500 km/s.

        Returns:
            :obj:`bool`: Flag whether or not the DAP can analyze the datacube.

        """
        return self.meta['vel'] is not None and self.meta['vel'] > -500

    def populate_metadata(self):
        r"""
        Populate and validate the :class:`DataCube` metadata (in :attr:`meta`)
        to ensure the it can be analyzed by the DAP.

        The only required metadata keyword is ``z``, which sets the initial
        guess for the bulk redshift of the galaxy.  If this key is not available
        or its value doesn't meet the criterion below, this function will raise
        an exception, meaning the DAP will fault before it starts processing the
        DataCube.

        The metadata provided must meet the following **critical criteria** or
        the method will fault:

            - Velocity (:math:`cz`) must be greater than -500.

        For the remainder of the metadata, if the keyword does not exist, if the
        value is None, or if the value is outside the accepted range, a default
        is chosen. The metadata keywords, acceptable ranges, and defaults are
        provided below.

        +-------------+--------------------------------+---------+
        |   Keyword   |                          Range | Default |
        +=============+================================+=========+
        | ``vdisp``   |             :math:`\sigma > 0` |     100 |
        +-------------+--------------------------------+---------+
        | ``ell``     | :math:`0 \leq \varepsilon < 1` |    None |
        +-------------+--------------------------------+---------+
        | ``pa``      |    :math:`0 \leq \phi_0 < 360` |    None |
        +-------------+--------------------------------+---------+
        | ``reff``    |        :math:`R_{\rm eff} > 0` |    None |
        +-------------+--------------------------------+---------+
        | ``ebv``     |          :math:`E(B-V) \geq 0` |    None |
        +-------------+--------------------------------+---------+
        | ``drpcrit`` |                  True or False |   False |
        +-------------+--------------------------------+---------+

        All other metadata in :attr:`meta` are ignored.

        If they do not already exist in :attr:`meta`, this method adds the
        following keywords:

            - ``z``: The bulk redshift of the galaxy, used to calculate
              :math:`cz`.

            - ``vel``: The initial guess velocity (:math:`cz`) in km/s.

            - ``vdisp``: The initial guess velocity dispersion in km/s.

            - ``ell``: The isophotal ellipticity (:math:`1-b/a`) to use when
                calculating semi-major axis coordinates.

            - ``pa``: The isophotal position angle in deg from N through E, used
              when calculating semi-major axis coordinates.

            - ``reff``: The effective radius in arcsec (DataCube WCS coordinates
              are expected to be in deg), used as a normalization of the
              semi-major axis radius in various output data.

            - ``ebv``: The E(B-V) Galactic reddening along the line-of-sight to
              the galaxy.

            - ``drpcrit``: A flag that indicates critical failures in the data
              reduction (True implies poor data quality).  In MaNGA, the DRP will
              flag datacubes as having CRITICAL failures, but the survey
              approach was to run the DAP on these datacubes anyway.  This flag
              is then used to propagate the caution in using the data analysis
              products to the user.
        """
        keys = self.meta.keys()
        if 'z' not in keys or self.meta['z'] is None:
            warnings.warn('The bulk redshift has not been defined.  The DAP will *not* be able '
                          'to analyze this datacube.')
            self.meta['z'] = None
        self.meta['vel'] = None if self.meta['z'] is None \
                            else astropy.constants.c.to('km/s').value * self.meta['z']

        if 'vdisp' not in keys or self.meta['vdisp'] is None or not self.meta['vdisp'] > 0:
            warnings.warn('Velocity dispersion not provided or not greater than 0 km/s.  '
                        'Adopting initial guess of 100 km/s.')
            self.meta['vdisp'] = 100.0

        if 'ell' not in keys or self.meta['ell'] is None or self.meta['ell'] < 0 \
                or self.meta['ell'] > 1:
            warnings.warn('Ellipticity not provided or not in the range 0 <= ell <= 1.  '
                        'Setting to None.')
            self.meta['ell'] = None

        if 'pa' not in keys or self.meta['pa'] is None:
            warnings.warn('Position angle not provided. Setting to None.')
            self.meta['pa'] = None

        # Impose expected range
        if self.meta['pa'] is not None and (self.meta['pa'] < 0 or self.meta['pa'] >= 360):
            warnings.warn('Imposing 0-360 range on position angle.')
            while self.meta['pa'] < 0:
                self.meta['pa'] += 360
            while self.meta['pa'] >= 360:
                self.meta['pa'] -= 360

        if 'reff' not in keys or self.meta['reff'] is None or not self.meta['reff'] > 0:
            warnings.warn('Effective radius not provided or not greater than 0. Setting to None.')
            self.meta['reff'] = None

        if 'ebv' not in keys or self.meta['ebv'] is None or self.meta['ebv'] < 0:
            warnings.warn('Galactic reddening not provided or not >= 0. Setting to None.')
            self.meta['ebv'] = None

        if 'drpcrit' not in keys or not isinstance(self.meta['drpcrit'], bool):
            warnings.warn('Data reduction quality flag not provided or not a boolean.  Setting '
                          'quality flag to False (i.e., data reduction is expected to be high '
                          'quality).')
            self.meta['drpcrit'] = False

        if not self.can_analyze:
            warnings.warn('The DAP will not be able to analyze this datacube.  See '
                          '"DataCube.can_analyze".')

    @classmethod
    def from_config(cls, cfgfile, **kwargs):
        """
        Construct a datacube object using a configuration file.

        The method is undefined for the base class. If called, an
        exception is raised. Derived classes must override this
        method to allow for a configuration file to be used to
        instantiate the relevant :class:`DataCube` subclass.

        Args:
            cfgfile (:obj:`str`):
                Configuration file. See `configparser.ConfigParser`_.
            **kwargs:
                Any other keyword arguments that invoke optional
                instantiation methods. Note that these arguments will
                *never* be used in a command-line level execution of
                the DAP. They should only be available for custom
                scripts.
        """
        raise NotImplementedError('No from_config method is available for {0}.'.format(
                                  cls.__name__))

    # TODO: write a from_rss classmethod

    def _get_pixelscale(self):
        """
        Measure the pixel scale using the WCS.

        The method uses the coordinates of 4 adjacent pixels to
        compute the pixel area. Assuming the pixels are square and
        that the pixel scale is constant in square arcsec over the
        full image, the square-root of the area is the pixel scale.

        The method will fault if :attr:`wcs` is None.

        Returns:
            :obj:`float`: The estimated pixel scale.
        """
        # TODO: Instead get the mean over the full image?
        coo = numpy.array([[1,1,2,2], [1,2,2,1], [1,1,1,1]]).T
        x, y, _ = self.wcs.all_pix2world(coo, 1).T
        x = (x - x[0])*numpy.cos(numpy.radians(y[0]))
        return numpy.sqrt(polygon_area(x, y))*3600

    def _get_wavelength_vector(self):
        """
        Use the `astropy.wcs.WCS`_ attribute (:attr:`wcs`) to
        generate the datacube wavelength vector.

        :attr:`wcs` cannot be None and the wavelength coordinate
        system must be defined along its third axis.

        Returns:
            `numpy.ndarray`_: Vector with wavelengths along the third
            axis defined by :attr:`wcs`.

        Raises:
            ValueError:
                Raised if :attr:`wcs` or :attr:`nwave` is not defined.
        """
        if self.nwave is None:
            raise ValueError('Length of the spectral axis (nwave) must be defined.')
        if self.wcs is None:
            raise ValueError('World coordinate system required to construct wavelength vector.')
        coo = numpy.array([numpy.ones(self.nwave), numpy.ones(self.nwave),
                           numpy.arange(self.nwave)+1]).T
        return self.wcs.all_pix2world(coo, 1)[:,2]*self.wcs.wcs.cunit[2].to('angstrom')

    @property
    def nspec(self):
        """Number of spectra in the datacube."""
        return numpy.prod(self.spatial_shape)

    @staticmethod
    def propagate_flags():
        """
        Flags that should be propagated from the observed data to the analyzed data.

        The base class returns ``None``.
        """
        return None

    @staticmethod
    def do_not_use_flags():
        """
        Return the maskbit names that should not be used.

        The base class returns ``None``.
        """
        return None

    @staticmethod
    def do_not_fit_flags():
        """
        Return the maskbit names that should not be fit.

        The base class returns ``None``.
        """
        return None

    @staticmethod
    def do_not_stack_flags():
        """
        Return the maskbit names that should not be stacked.
        
        The base class returns ``None``.
        """
        return None

    def metakeys(self):
        """Get a :obj:`list` of the keys in the datacube metadata."""
        return list(self.meta.keys())

    # TODO: Add a getitem method that returns the datacube flux?

    @property
    def can_compute_covariance(self):
        """
        Determine if the object can be used to compute the spatial
        covariance.

        If :attr:`covar` is currently not defined, the method tries
        to load the row-stacked spectra used to build the datacube;
        see :func:`load_rss`. If that is successful or if
        :attr:`covar` is already defined, the method returns True. If
        :attr:`covar` is None and the row-stacked spectra cannot be
        loaded, the method returns False.

        Returns:
            :obj:`bool`: Flag that the object can be used to
            calculate the spatial covariance.
        """
        if self.covar is None:
            self.load_rss()
        if self.covar is None and self.rss is None:
            return False
        return True

    def load_rss(self, **kwargs):
        """
        Try to load the source row-stacked spectra for this datacube.

        This method is undefined in the base class, and simply passes.

        Derived classes should override this method if it's possible
        to load the row-stacked spectra. If the load is successful,
        :attr:`rss` should no longer be None.
        """
        pass

    def copy_to_array(self, attr='flux', waverange=None, nbins=None, select_bins=None,
                      missing_bins=None, unique_bins=None):
        r"""
        Return a copy of the selected data array with a flattened
        spatial axis.

        The array size is always :math:`N_{\rm spec} \times N_{\rm
        wavelength}`. The spatial positions within the original
        datacube for each spectrum are given by tuples in
        :attr:`spatial_index`.

        See :func:`copy_to_masked_array` for argument descriptions.

        .. warning::
            Any masking is ignored in this function call. To
            incorporate the mask use :func:`copy_to_masked_array`.

        Returns:
            `numpy.ndarray`_: A 2D array with a copy of the data from the
            selected attribute.
        """
        masked_data = self.copy_to_masked_array(attr=attr, use_mask=False, waverange=waverange,
                                                nbins=nbins, select_bins=select_bins,
                                                missing_bins=missing_bins, unique_bins=unique_bins)
        # For this approach, the wavelengths masked should be
        # *identical* for all spectra
        nwave = numpy.sum(numpy.invert(numpy.ma.getmaskarray(masked_data)), axis=1)
        # No masking should be present except for the wavelength range
        # meaning that the number of unmasked pixels at all spatial
        # positions should be the same.
        if numpy.any(nwave != nwave[0]):
            raise ValueError('Masking in copy_to_mask should only for the wavelength range.')
        if numpy.all(nwave == 0):
            raise ValueError('Full wavelength range has been masked!')
        # Compression returns a flattened array, so it needs to be
        # reshaped for the new number of (unmasked) wavelength channels
        return masked_data.compressed().reshape(-1,nwave[0])

    def copy_to_masked_array(self, attr='flux', use_mask=True, flag=None, waverange=None,
                             nbins=None, select_bins=None, missing_bins=None, unique_bins=None):
        r"""
        Return a copy of the selected data array as a masked array
        with a flattened spatial axis.

        The array size is always :math:`N_{\rm spec} \times N_{\rm
        wavelength}`. The spatial positions within the original
        datacube for each spectrum are given by tuples in
        :attr:`spatial_index`.

        This is functionally identical to :func:`copy_to_array`,
        except the output format is a `numpy.ma.MaskedArray`_. The
        pixels that are considered to be masked can be specified
        using the `flag` option.
        
        Args:
            attr (:obj:`str`, optional):
                The attribute for the returned array. Can be 'flux',
                'ivar', 'sres', or 'mask'. For 'mask', you're likely
                better off using :func:`copy_to_array`. Strings are
                always set to be lower-case, so capitalization
                shouldn't matter.
            use_mask (:obj:`bool`, optional):
                Use the internal mask to mask the data. This is
                largely here to allow for :func:`copy_to_array` to
                wrap this function while not applying the internal
                mask.
            waverange (array-like, optional):
                Two-element array with the first and last wavelength
                to include in the computation. Default is to use the
                full wavelength range.
            flag (:obj:`str`, :obj:`list`, optional):
                One or more bitmask flags used to select bad pixels.
                The names *must* be a valid bit name as defined by
                :attr:`bitmask` (see
                :class:`mangadap.util.bitmask.BitMask`). If not
                provided, *any* non-zero mask bit is omitted.
            nbins (:obj:`int`, optional):
                The total number of defined bins. Default is to find
                the maximum number in the unique_bins list.
            select_bins (array-like, optional):
                A boolean array selecting spectra in the flattened
                cube (with shape :math:`N_{\rm spec} \times
                N_\lambda`) to return.
            missing_bins (:obj:`list`, optional):
                A list of bin numbers that are missing from the
                output selection and that will be replaced with
                masked spectra. If specified, must also provide
                ``unique_bins``.
            unique_bins (array-like, optional):
                The indices of the bins that have valid spectra. The
                length of this array should match the number of
                selected spectra from ``select_bins``.

        Returns:
            `numpy.ma.MaskedArray`_: A 2D array with a copy of the
            data from the selected extension, masked where
            :attr:`mask` is either True or with the selected bitmask
            flags.

        Raises:
            AttributeError:
                Raised if ``attr`` is not a valid attribute.
            ValueError:
                Raised if the unique bins are not specified, but the
                missing bins are, or if the number of unique bins
                does not match the number of spectra in the datacube.
        """
        # Make sure masks and flags make sense
        if flag is not None and self.bitmask is None:
            warnings.warn('No bitmask defined.  Input flags ignored.')

        nspec = self.nspec

        # Create the wavelength mask (will be all true if
        # waverange=None)
        mask = SpectralPixelMask(waverange=waverange).boolean(self.wave, nspec=nspec)

        # Add in any masked data
        if use_mask:
            mask |= self.mask.reshape(nspec,-1) if self.bitmask is None \
                    else self.bitmask.flagged(self.mask.reshape(nspec,-1), flag=flag)

        # Create the output MaskedArray
        # TODO: Handle when attr is None
        a = numpy.ma.MaskedArray(getattr(self, attr.lower()).reshape(nspec,-1), mask=mask)

        # Apply any bin selection
        if select_bins is not None:
            a = a[select_bins,:]

        # No missing bins are specified, so return the array
        if missing_bins is None or len(missing_bins) == 0:
            return a

        # Missing bins have been specified, so the unique_bins also need
        # to be specified
        if unique_bins is None:
            raise ValueError('Must specify the unique bins if missing bins are specified.')
        if len(unique_bins) != a.shape[0]:
            raise ValueError('Number of unique bins does not match the number of spectra.')

        # If the number of bins hasn't been provided, construct it based
        # on the unique bins specified
        _nbins = max(unique_bins)+1 if nbins is None else nbins

        # Create the array with the right shape and mask everything
        _a = numpy.ma.MaskedArray(numpy.zeros((_nbins, a.shape[-1]), dtype=a.dtype),
                                  mask=numpy.ones((_nbins, a.shape[-1]), dtype=bool))
        # Copy the data to the correct bins, which also unmasks these
        # pixels
        _a[unique_bins,:] = a
        # Return the masked array; "missing bins" are fully masked
        return _a

    def interpolate_to_match(self, func, fill_value=0.0):
        r"""
        Interpolate a function to match the datacube wavelength sampling.

        Args:
            func (`numpy.ndarray`_):
                Function to linear interpolate. Shape is
                :math:`(N_{\rm wave}, 2)`, where the first column has
                the wavelength and the second has the function to
                interpolate. If None, simply returns a unity vector
                of the correct length.
            fill_value (:obj:`float`, optional):
                The value to use for spectral regions not sampled by
                the provided function.

        Returns:
            `numpy.ndarray`_: The function interpolated (**not**
            resampled) to match :attr:`wave`. Any spectral regions
            not within the wavelength range of the provided function
            is set to 0.
        """
        if func is None:
            return numpy.ones(self.nwave, dtype=float)
        return interpolate.interp1d(func[:,0], func[:,1], bounds_error=False,
                                    fill_value=fill_value, assume_sorted=True)(self.wave)

    def mean_sky_coordinates(self, center_coo=None):
        r"""
        Compute the mean sky coordinates for each spectrum.

        If the WCS is available, the coordinates can be returned in
        RA and declination (degrees). If center coordinates are
        provided (see ``center_coo``), however, the coordinates are
        offset set as follows:

        .. math::

            x &= (\alpha - \alpha_0) \cos \delta_0 \\
            y &= (\delta - \delta_0),

        where :math:`(\alpha_0, \delta_0)` are the provided
        coordinates.

        If the WCS is not available, the returned coordinates are in
        arcsec from the center of the image (regardless of the value
        of ``center_coo``) determined using the pixelscale. At least
        in this case, the coordinates are assumed to be relative to
        the pixel center (not, e.g., its edge).

        Args:
            center_coo (:obj:`tuple`, optional):
                A two-tuple with the coordinates in right-ascension
                and declination for the coordinate-frame origin. If
                None, no offset is performed.

        Returns:
            :obj:`tuple`: Two `numpy.ndarray`_ objects with the RA
            and declination of each pixel in degrees, or its offset
            from the center in arcseconds. In both cases the shape of
            the returned arrays matches the spatial shape of the
            datacube.
        """
        i, j = numpy.meshgrid(numpy.arange(self.spatial_shape[0]),
                              numpy.arange(self.spatial_shape[1]), indexing='ij')
        if self.wcs is None:
            return (i-self.spatial_shape[0]//2+(self.spatial_shape[0]%2)*0.5)*self.pixelscale, \
                   (j-self.spatial_shape[1]//2+(self.spatial_shape[1]%2)*0.5)*self.pixelscale

        # Generate pixel coordinates to pass to the WCS; just use the
        # first wavelength channel.
        coo = numpy.array([i.ravel()+1, j.ravel()+1, numpy.ones(self.nspec)]).T
        x, y, _ = [c.reshape(self.spatial_shape) for c in self.wcs.all_pix2world(coo, 1).T]
        if center_coo is None:
            # Not offsetting, so we're done
            return x, y

        # TODO: Use astropy.coordinates.SkyOffset 
        # Offset and return
        if len(center_coo) != 2:
            raise ValueError('Provided coordinates are expected to be a single x,y pair.')
        return (x-center_coo[0]) * numpy.cos(numpy.radians(center_coo[1])) * 3600., \
                    (y-center_coo[1]) * 3600.

    def binned_on_sky_area(self, bin_indx):
        r"""
        Compute the on-sky area of a set of binned spectra.

        For each bin, this is just the number of spaxels in the bin
        multiplied by the spaxel area.

        Args:
            bin_indx (array-like):
                An array with size :math:`N_{\rm spec}` that gives
                which spaxels were included in each bin. Valid bins
                have indices of :math:`\geq 0`.

        Returns:
            :obj:`tuple`: Two `numpy.ndarray`_ objects are returned.
            The first has the unique (non-negative) bin indices, and
            the second provides the on-sky area of that bin.
        """
        unique_bins, bin_count = numpy.unique(bin_indx, return_counts=True)
        indx = unique_bins > -1
        return unique_bins[indx], (bin_count[indx]*numpy.square(self.pixelscale)).astype(float)

    def central_wavelength(self, waverange=None, response_func=None, per_pixel=True, flag=None,
                           fluxwgt=False):
        """
        Determine the mean central wavelength for all spectra under
        various conditions.

        The wavelength channel is set to be the center of the
        bandpass selected by ``waverange``, weighted by the response
        function and the flux (if requested/provided). The mask is
        also incorporated in these calculations. By default (i.e., no
        wavelength limits or weighting) the wavelength channel is
        just the central wavelength of the full spectral range. (Note
        that if the spectra are binned logarithmically, this isn't
        necessarily the central wavelength *channel*.)

        Args:
            waverange (array-like, optional):
                Starting and ending wavelength over which to
                calculate the statistics. Default is to use the full
                wavelength range.
            response_func (array-like, optional):
                A two-column array with the wavelength and
                transmission of a broad-band response function to use
                as a weighting function for the calculation.
            per_pixel (:obj:`bool`, optional):
                When providing a response function, base the
                calculation on per pixel measurements, instead of per
                angstrom. Set to False for a per-angstrom
                calculation.
            flag (:obj:`str`, :obj:`list`, optional):
                One or more flag names that are considered when
                deciding if a pixel should be masked. The names
                *must* be a valid bit name as defined by
                :attr:`bitmask`. If :attr:`bitmask` is None, these
                are ignored.
            fluxwgt (:obj:`bool`, optional):
                Flag to weight by the flux when determining the mean
                coordinates.

        Returns:
            :obj:`float`: The mean central wavelength of all spectra.
        """
        if waverange is None and response_func is None and not fluxwgt:
            return (self.wave[0] + self.wave[-1])/2.

        flux = self.copy_to_masked_array(waverange=waverange, flag=flag)
        dw = numpy.ones(self.nwave, dtype=float) if per_pixel \
                else angstroms_per_pixel(self.wave, log=self.log)
        _response_func = self.interpolate_to_match(response_func)

        if fluxwgt:
            norm = numpy.ma.sum(flux*_response_func[None,:]*dw[None,:], axis=1)
            cen_wave = numpy.ma.sum(flux*self.wave[None,:]*_response_func[None,:]*dw[None,:],
                                    axis=1) / norm
            return float(numpy.ma.mean(cen_wave))

        norm = numpy.ma.sum(numpy.invert(numpy.ma.getmaskarray(flux))
                            * _response_func[None,:] * dw[None,:], axis=1)
        cen_wave = numpy.ma.sum(numpy.invert(numpy.ma.getmaskarray(flux)) * self.wave[None,:]
                                * _response_func[None,:] * dw[None,:], axis=1) / norm
        return float(numpy.ma.mean(cen_wave))

    def flux_stats(self, waverange=None, response_func=None, per_pixel=True, flag=None):
        r"""
        Compute the mean flux, propagated error in the mean flux, and
        mean S/N over the specified wavelength range.

        If the wavelength range is not specified, the quantities are
        calculated over the full spectral range.

        Args:
            waverange (array-like, optional):
                Starting and ending wavelength over which to
                calculate the statistics. Default is to use the full
                wavelength range.
            response_func (array-like, optional):
                A two-column array with the wavelength and
                transmission of a broad-band response function to use
                as a weighting function for the calculation.
            per_pixel (:obj:`bool`, optional):
                When providing a response function, continue to
                calculate the statistics per pixel. Set to False for
                a per-angstrom calculation.
            flag (:obj:`str`, :obj:`list`, optional):
                One or more flag names that are considered when
                deciding if a pixel should be masked. The names
                *must* be a valid bit name as defined by
                :attr:`bitmask`.

        Returns:
            `numpy.ndarray`_: Three objects are returned: the mean
            flux, the propagated variance in the mean flux, and the
            mean S/N. The shape of each is the same as the shape of a
            single wavelength channel in the datacube (i.e.,
            :attr:`spatial_shape`).
    
        Raises:
            ValueError:
                Raised of a provided wavelength range object does not
                have two elements or if a response function is
                provided and has an incorrect shape.
        """
        if waverange is not None and len(waverange) != 2:
            raise ValueError('Provided wavelength range must be a two-element vector.')
        if response_func is not None:
            if len(response_func.shape) != 2:
                raise ValueError('Response function object must be two dimensional.')
            if response_func.shape[1] != 2:
                raise ValueError('Response function object must only have two columns.')

        # Grab the masked arrays
        flux = self.copy_to_masked_array(waverange=waverange, flag=flag)
        ivar = self.copy_to_masked_array(attr='ivar', waverange=waverange, flag=flag)
        snr = flux*numpy.ma.sqrt(ivar)

        # Set the response function
        dw = numpy.ones(self.nwave, dtype=float) if per_pixel \
                else angstroms_per_pixel(self.wave, log=self.log)
        _response_func = self.interpolate_to_match(response_func)

        # Calculate the statistics and return
        response_integral = numpy.sum(numpy.invert(numpy.ma.getmaskarray(flux))
                                        * (_response_func*dw)[None,:], axis=1)
        signal = numpy.ma.divide(numpy.ma.sum(flux*(_response_func*dw)[None,:], axis=1),
                                 response_integral).reshape(self.spatial_shape)
        variance = numpy.ma.divide(numpy.ma.sum(numpy.ma.power(ivar, -1.) \
                                    * (_response_func*dw)[None,:], axis=1),
                                   response_integral).reshape(self.spatial_shape)
        snr = numpy.ma.divide(numpy.ma.sum(snr*(_response_func*dw)[None,:], axis=1),
                              response_integral).reshape(self.spatial_shape)
        return signal, variance, snr

    def covariance_matrix(self, channel, **kwargs):
        """
        Construct the formal covariance matrix for the provided
        channel.

        This is a simple wrapper for a call to
        :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.covariance_matrix`
        executed using :attr:`rss`, which cannot be None. See that
        method for the argument description.

        .. warning::

            The provided rectifications parameters should be the same
            as used to construct the datacube. If not, the calculated
            covariance matrix will not be correct.

        """
        if self.rss is None:
            raise ValueError('RowStackedSpectra object not available for this datacube!')
        return self.rss.covariance_matrix(channel, **kwargs)

    def covariance_cube(self, **kwargs):
        """
        Construct the formal covariance matrix for one or more
        channels.

        This is a simple wrapper for a call to
        :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.covariance_cube`
        executed using :attr:`rss`, which cannot be None. See that
        method for the argument description.

        .. warning::

            The provided rectifications parameters should be the same
            as used to construct the datacube. If not, the calculated
            covariance matrix will not be correct.

        """
        if self.rss is None:
            raise ValueError('RowStackedSpectra object not available for this datacube!')
        return self.rss.covariance_cube(**kwargs)

    def approximate_correlation_matrix(self, sigma_rho, rlim, rho_tol=None, redo=False):
        r"""
        Construct a correlation matrix with correlation coefficients
        that follow a Gaussian in 2D pixel separation.

        This method constructs a correlation matrix with correlation
        coefficients defined as

        .. math::

            \rho_{ij} = \exp(-D_{ij}^2 / 2\sigma_{\rho}^2)

        where :math:`D_{ij}` is the distance between two spaxels in
        the spatial dimension of the datacube (in the number spaxels,
        not arcsec). Any pixels with :math:`D_{ij} > R_{\rm lim}` is
        set to zero, where :math:`R_{\rm lim}` is the limiting radius
        of the kernel used in the datacube rectification; this is
        provided here as ``rlim`` in arcsec, which is converted to
        spaxels using :attr:`pixelscale`.

        We found in Westfall et al. (2019, AJ, 158, 231) that this is
        a reasonable approximation for the formal covariance that
        results from Shepard's rectification method.

        There is an unknown relation between the dispersion of the
        kernel used by Shepard's method and the value of
        :math:`\sigma_{\rho}`. Tests show that it is tantalizingly
        close to :math:`\sigma_{\rho} = \sqrt{2}\sigma`, where
        :math:`\sigma` is in pixels instead of arcsec; however, a
        formal derivation of this hasn't been done and is complicated
        by the focal-plane sampling of the row-stacked spectra.

        Within the limits of how the focal-plane sampling changes
        with wavelength, we found the value of :math:`\sigma_{\rho}`
        varies little with wavelength in MaNGA. Here, we assume
        :math:`\sigma_{\rho}` is fully wavelength independent.
        Therefore, once this method is run once, it doesn't need to
        be run again for different wavelength channels. To force the
        correlation matrix to be recreated, use ``redo`` or change
        the provided :math:`\sigma_{\rho}`.

        Args:
            sigma_rho (:obj:`float`):
                The :math:`\sigma_{\rho}` of the Gaussian function
                used to approximate the trend of the correlation
                coefficient with spaxel separation.
            rlim (:obj:`float`):
                The limiting radius of the image reconstruction
                (Gaussian) kernel in arcseconds.
            rho_tol (:obj:`float`, optional):
                Any correlation coefficient less than this is assumed
                to be equivalent to (and set to) 0.
            redo (:obj:`bool`, optional):
                Force the recalculation of the cube dimensions if
                they are already defined and :math:`\sigma_{\rho}` has
                not changed.

        Returns:
            :class:`mangadap.util.covariance.Covariance`: Correlation
            matrix
        """
        if self.approx_correl is not None \
                and any([sigma_rho == this for this in [None, self.sigma_rho]]) \
                and any([rlim == this for this in [None, self.correl_rlim]]):
            return self.approx_correl

        if sigma_rho is None or rlim is None:
            raise ValueError('Must provide sigma_rho and rlim for approximate correlation '
                             'matrix calculation.')

        # Get the full covariance grid
        ii, jj = map(lambda x: x.ravel(),
                     numpy.meshgrid(numpy.arange(self.nspec), numpy.arange(self.nspec),
                                    indexing='ij'))

        # Convert covariance pixels to spatial pixels along both dimensions
        i_i, i_j = numpy.unravel_index(ii, self.spatial_shape)
        j_i, j_j = numpy.unravel_index(jj, self.spatial_shape)

        # Get the (square of the) distances from each spaxel to every
        # other spaxel
        dij = numpy.square(j_i-i_i) + numpy.square(j_j-i_j)
        indx = dij <= numpy.square(2*rlim/self.pixelscale)

        # Calculate the correlation coefficient
        ii = ii[indx]
        jj = jj[indx]
        rhoij = numpy.exp(-dij[indx]/numpy.square(sigma_rho)/2)

        if rho_tol is not None:
            indx = rhoij > rho_tol
            ii = ii[indx]
            jj = jj[indx]
            rhoij = rhoij[indx]

        # Construct the Covariance object and save the input
        self.sigma_rho = sigma_rho
        self.correl_rlim = rlim
        self.approx_correl = Covariance(sparse.coo_matrix((rhoij, (ii,jj)),
                                                          shape=(self.nspec,self.nspec)).tocsr(),
                                        impose_triu=True, correlation=True,
                                        raw_shape=self.spatial_shape)
        return self.approx_correl

    def approximate_covariance_matrix(self, channel, sigma_rho=None, rlim=None, rho_tol=None,
                                      csr=False, quiet=False):
        r"""
        Return an approximate calculation of the covariance matrix
        assuming
        
        .. math::

            C_{ij} = \rho_{ij}(V_{i} V_{j})^{1/2}

        where :math:`\rho_{ij}` is approximated by
        :func:`approximate_correlation_matrix` and :math:`V_i\equiv C_{ii}` are
        the variances provided by the inverse of :attr:`ivar`.

        The method first calculates :math:`\rho_{ij}` if it hasn't
        been yet or the provided ``sigma_rho`` and/or ``rlim`` values
        are different to previous calls, which builds
        :attr:`approx_correl`. If :attr:`approx_correl` is not yet
        constructed, ``sigma_rho`` and ``rlim`` must be provided.

        The returned covariance matrix is the correlation matrix
        rescaled by the variance provided by :attr:`ivar` for the
        specified channel.
        
        Args:
            channel (:obj:`int`):
                Index of the spectral channel for which to calculate
                the covariance matrix.
            sigma_rho (:obj:`float`, optional):
                The :math:`\sigma_{\rho}` of the Gaussian function
                used to approximate the trend of the correlation
                coefficient with spaxel separation.
            rlim (:obj:`float`, optional):
                The limiting radius of the image reconstruction
                (Gaussian) kernel in arcseconds.
            rho_tol (:obj:`float`, optional):
                Any correlation coefficient less than this is assumed
                to be equivalent to (and set to) 0.
            csr (:obj:`bool`, optional):
                Instead of returning a
                :class:`mangadap.util.covariance.Covariance` object,
                return the covariance matrix as a
                `scipy.sparse.csr_matrix`_ object. Primarily used by
                :func:`approximate_covariance_cube` for collating the
                covariance matrix of each wavelength channel before
                combining them into a single
                :class:`mangadap.util.covariance.Covariance` object.
            quiet (:obj:`bool`, optional):
                Suppress terminal output

        Returns:
            :class:`mangadap.util.covariance.Covariance`,
            `scipy.sparse.csr_matrix`_: The approximate covariance
            matrix for the designated wavelength channel. The return
            type depends on `csr`.
        """
        self.approximate_correlation_matrix(sigma_rho, rlim, rho_tol=rho_tol)
        var = numpy.ma.power(self.ivar[...,channel], -1).filled(0.0).ravel()
        covar = self.approx_correl.apply_new_variance(var)
        covar.revert_correlation()
        return covar.full() if csr else covar

    def approximate_covariance_cube(self, channels=None, sigma_rho=None, rlim=None, rho_tol=None,
                                    csr=False, quiet=False):
        r"""
        Return the approximate covariance matrices for many
        wavelength channels.

        This is a simple wrapper for
        :func:`approximate_covariance_matrix` that iteratively builds
        the covariance matrix for each wavelength channel.

        If :attr:`approx_correl` is not yet constructed (see
        :func:`approximate_correlation_matrix`), ``sigma_rho`` and
        ``rlim`` must be provided.
        
        Args:
            channels (:obj:`int`, array-like, optional):
                Indices of the spectral channels for which to
                calculate the covariance matrix. If None, the
                covariance matrix is calculated for *all* channels.
            sigma_rho (:obj:`float`, optional):
                The :math:`\sigma_{\rho}` of the Gaussian function
                used to approximate the trend of the correlation
                coefficient with spaxel separation.
            rlim (:obj:`float`, optional):
                The limiting radius of the image reconstruction
                (Gaussian) kernel in arcseconds.
            rho_tol (:obj:`float`, optional):
                Any correlation coefficient less than this is assumed
                to be equivalent to (and set to) 0.
            csr (:obj:`bool`, optional):
                Instead of returning a
                :class:`mangadap.util.covariance.Covariance` object,
                return the covariance matrix as a
                `scipy.sparse.csr_matrix`_ object. Primarily used by
                :func:`approximate_covariance_cube` for collating the
                covariance matrix of each wavelength channel before
                combining them into a single
                :class:`mangadap.util.covariance.Covariance` object.
            quiet (:obj:`bool`, optional):
                Suppress terminal output

        Returns:
            :class:`mangadap.util.covariance.Covariance`,
            `numpy.ndarray`_: If ``csr`` is True, the returned object
            is an `numpy.ndarray`_ of `scipy.sparse.csr_matrix`_
            types.
        """
        # Set the channels
        _channels = numpy.arange(self.nwave) if channels is None \
                        else numpy.atleast_1d(channels)
        nc = len(_channels)

        # Build and return the covariance matrices
        CovCube = numpy.empty(nc, dtype=sparse.csr.csr_matrix)
        for i in range(nc):
            if not quiet:
                print('Calculating covariance matrix: {0}/{1}'.format(i+1,nc), end='\r')
            CovCube[i] = self.approximate_covariance_matrix(_channels[i], sigma_rho=sigma_rho,
                                                            rlim=rlim, rho_tol=rho_tol, csr=True,
                                                            quiet=True)
        if not quiet:
            print('Calculating covariance matrix: {0}/{0}'.format(nc))
        return Covariance(CovCube, input_indx=_channels) if not csr else CovCube





import pytest

from IPython import embed

import numpy
from astropy.io import fits

from mangadap.config.manga import MaNGAConfig
from mangadap.proc.reductionassessments import ReductionAssessmentDef
from mangadap.util.covariance import Covariance
from mangadap.datacube import MaNGADataCube
from mangadap.tests.util import data_test_file, remote_data_file, requires_remote


@requires_remote
def test_sres_ext():
    cfg = MaNGAConfig(7815, 3702)
    file = remote_data_file(filename=cfg.file_name)
    hdu = fits.open(file)
    assert MaNGAConfig.spectral_resolution_extension(hdu) == 'LSFPRE', \
                'Bad spectral resolution extension selection'
    assert MaNGAConfig.spectral_resolution_extension(hdu, ext='SPECRES') == 'SPECRES', \
                'Bad spectral resolution extension selection'
    assert MaNGAConfig.spectral_resolution_extension(hdu, ext='junk') is None, \
                'Should return None for a bad extension name.'


@requires_remote
def test_read():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file())

    assert cube.log, 'Should read the log-binned version by default.'
    assert cube.wcs is not None, 'WCS should be defined.'
    assert cube.shape[:2] == cube.spatial_shape, 'Spatial shape should be first two axes.'
    assert cube.nspec == numpy.prod(cube.spatial_shape), 'Definition of number of spectra changed.'
    assert cube.sres is not None, 'Spectral resolution data was not constructed.'
    assert cube.sres_ext == 'LSFPRE', 'Should default to LSFPRE extension.'
    assert abs(cube.pixelscale - cube._get_pixelscale()) < 1e-6, 'Bad match in pixel scale.'
    # NOTE: This is worse than it should be because of how the WCS in MaNGA is defined.
    assert numpy.all(numpy.absolute(cube.wave - cube._get_wavelength_vector()) < 2e-4), \
            'Bad calculation of wavelength vector.'
    assert cube.covar is None, 'Covariance should not have been read'


@requires_remote
def test_read_correl():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file(),
                                       covar_ext='GCORREL')
    assert isinstance(cube.covar, Covariance), 'Incorrect type for covariance.'
    assert cube.covar.shape == (cube.nspec,cube.nspec), 'Covariance has incorrect shape.'
    assert cube.covar.is_correlation, 'Covariance object should be in a correlation mode.'
    
    # Check that the variances are all unity (or close to it when it's defined)
    unique_var = numpy.unique(cube.covar.var)
    assert numpy.allclose(unique_var[unique_var>0], 1.), 'Bad variance values'


@requires_remote
def test_wcs():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file())
    x, y = cube.mean_sky_coordinates(offset=None)
    assert x[0,0] > x[-1,0], 'RA should increase from large to small indices'
    assert y[0,0] < y[0,-1], 'DEC should increase from small to small indices'
    assert numpy.unravel_index(numpy.argmin( numpy.square(x - cube.prihdr['OBJRA']) 
                                            + numpy.square(y - cube.prihdr['OBJDEC'])), x.shape) \
                == (21,21), 'Object should be at cube center.'
    x, y = cube.mean_sky_coordinates(center_coo=(x[0,0], y[0,0]))
    assert numpy.isclose(x[0,0], 0.0) and numpy.isclose(y[0,0], 0.0), 'Offset incorrect'
    x, y = cube.mean_sky_coordinates()
    assert abs(x[21,21]) < 1e-2 and abs(y[21,21]) < 1e-2, 'Offset incorrect'


@requires_remote
def test_copyto():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file())
    flux = cube.copy_to_array()
    assert not isinstance(flux, numpy.ma.MaskedArray), 'Should output normal array'
    assert flux.shape[0] == cube.nspec, 'Should be flattened into a 2D array.'
    assert flux.shape[1] == cube.nwave, 'Should be flattened into a 2D array.'

    # Apply a wavelength mask
    waverange = [5000, 7000]
    flux = cube.copy_to_array(waverange=waverange)
    indx = (cube.wave > waverange[0]) & (cube.wave < waverange[1])
    assert flux.shape[1] == numpy.sum(indx), 'Wavelength range masking failed'

    # Find the spaxels with non-zero signal
    method = ReductionAssessmentDef()
#    methods = available_reduction_assessments()
#    i = numpy.where([m['key'] == 'SNRG' for m in methods])[0]
#    assert len(i) == 1, 'Could not find correct reduction assessment definition.'
    sig, var, snr = cube.flux_stats(response_func=method.response)
    indx = ((sig > 0) & numpy.invert(numpy.ma.getmaskarray(sig))).data.ravel()
    ngood = numpy.sum(indx)

    # Select the spaxels with non-zero signal
    flux = cube.copy_to_array(waverange=waverange, select_bins=indx)
    assert flux.shape[0] == ngood, 'Bin selection failed'

    # Get the masked array
    flux = cube.copy_to_masked_array()
    assert isinstance(flux, numpy.ma.MaskedArray), 'Should output a masked array'
    assert flux.shape[0] == cube.nspec, 'Should be flattened into a 2D array.'
    assert flux.shape[1] == cube.nwave, 'Should be flattened into a 2D array.'

    # Select the spaxels with non-zero signal
    flux = cube.copy_to_masked_array(select_bins=indx)
    assert flux.shape[0] == ngood, 'Bin selection failed'

    # Try to get the inverse variance
    i = cube.nspec//2 + cube.spatial_shape[1]//2
    ivar = cube.copy_to_masked_array(attr='ivar')
    assert ivar.shape == (cube.nspec, cube.nwave), 'Bad ivar shape'
    assert numpy.array_equal(cube.ivar[numpy.unravel_index(i, cube.spatial_shape)],
                             ivar[i].data), 'Did not pull ivar data.'

    # Try to get the spectral resolution
    sres = cube.copy_to_masked_array(attr='sres')
    assert sres.shape == (cube.nspec, cube.nwave), 'Bad sres shape'
    assert numpy.array_equal(cube.sres[numpy.unravel_index(i, cube.spatial_shape)],
                             sres[i].data), 'Did not pull sres data.'

@requires_remote
def test_stats():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file())

    # Create a fake bin map
    bin_indx = numpy.arange(cube.nspec/4, dtype=int).reshape(cube.spatial_shape[0]//2,
                                                             cube.spatial_shape[0]//2)
    bin_indx = numpy.repeat(bin_indx, 2, axis=0)
    bin_indx = numpy.repeat(bin_indx, 2, axis=1)

    # Get the bin area
    bins, area = cube.binned_on_sky_area(bin_indx)

    assert numpy.array_equal(bins, numpy.arange(cube.nspec/4)), 'Bad bin list'
    assert numpy.allclose(area, 1.), 'Bad area calculation'

    # Find the spaxels with non-zero signal
    method = ReductionAssessmentDef()
#    methods = available_reduction_assessments()
#    i = numpy.where([m['key'] == 'SNRG' for m in methods])[0]
#    assert len(i) == 1, 'Could not find correct reduction assessment definition.'
    cen_wave = cube.central_wavelength(response_func=method.response,
                                       flag=cube.do_not_use_flags())
    assert numpy.isclose(cen_wave, 4638.0), 'Central wavelength changed.'

    cen_wave = cube.central_wavelength(waverange=[4000,8000], flag=cube.do_not_use_flags(),
                                       fluxwgt=True)
    assert numpy.isclose(cen_wave, 5895.7), 'Central wavelength changed.'

    cen_wave = cube.central_wavelength(waverange=[4000,8000], flag=cube.do_not_use_flags(),
                                       per_pixel=False)
    assert numpy.isclose(cen_wave, 6044.9), 'Central wavelength changed.'

    sig, var, snr = cube.flux_stats(response_func=method.response)
    assert sig.shape == cube.spatial_shape, 'Should be shaped as a map.'
    assert isinstance(sig, numpy.ma.MaskedArray), 'Expected masked arrays'
    assert numpy.ma.amax(snr) > 60, 'S/N changed'

    # Try it with the linear cube
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file(), log=False)
    _sig, _var, _snr = cube.flux_stats(response_func=method.response)
    # TODO: Not sure why these are not closer.
    assert numpy.absolute(numpy.ma.median((sig-_sig)/_sig)) < 0.01, \
            'Signal should be the same to better than 1%.'
    assert numpy.absolute(numpy.ma.median((var-_var)/_var)) < 0.03, \
            'Variance should be the same to better than 3%.'
    assert numpy.absolute(numpy.ma.median((snr-_snr)/_snr)) < 0.02, \
            'S/N should be the same to better than 2%.'


@requires_remote
def test_read_lin():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file(), log=False)
    assert not cube.log, 'Wavelength sampling should be linear'
    assert numpy.isclose(numpy.std(numpy.diff(cube.wave)), 0.), \
                'Wavelength sampling should be linear'


@requires_remote
def test_from_config():
    cube = MaNGADataCube.from_config(data_test_file('datacube.ini'))
    assert cube.meta['z'] == 0.0293823, 'Bad config file read'
    assert cube.meta['ell'] == 0.110844, 'Bad config file read'


@requires_remote
def test_load_rss():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file())
    cube.load_rss()


@requires_remote
def test_covariance():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file())

    with pytest.raises(ValueError):
        # Have to load the RSS first
        cube.covariance_matrix(1000)

    # Load the RSS
    cube.load_rss()

    # Construct a covariance matrix
    C = cube.covariance_matrix(1000)
    assert C.shape == (1764, 1764), 'Bad covariance shape'

    # Make it a correlation matrix and check it
    C.to_correlation()

    # Check that the variances are all unity (or close to it when it's defined)
    unique_var = numpy.unique(numpy.diag(C.toarray()))
    assert numpy.allclose(unique_var[unique_var>0], 1.), 'Bad correlation diagonal'

    # Try multiple channels
    C = cube.covariance_cube(channels=[1000,2000])
    assert numpy.array_equal(C.input_indx, [1000,2000]), 'Bad matrix indices'
    assert C.shape == (1764, 1764, 2), 'Bad covariance shape'

    # Try to convert multiple channels
    C.to_correlation()
    # And reverting it
    C.revert_correlation()

    # Try to generate an approximate correlation matrix, covariance
    # matrix, and covariance cube
    approxC = cube.approximate_correlation_matrix()
    approxC = cube.approximate_covariance_matrix(1000)
    approxC = cube.approximate_covariance_cube(channels=[1000,2000])

    # Variance should be the same for direct and approximate calculations
    assert numpy.allclose(approxC.variance(), C.variance()), 'Variances should be the same.'




from IPython import embed

from mangadap.par.emissionlinedb import EmissionLineDB


def test_read():
    dbs = EmissionLineDB.available_databases()
    assert len(dbs) > 0, 'No emission-line databases available'
    for key in dbs.keys():
        emldb = EmissionLineDB.from_key(key)


def test_mpl11():
    emldb = EmissionLineDB.from_key('ELPMPL11')
    assert len(emldb) == 35, 'Incorrect number of emission lines'
    assert 'ArIII' in emldb['name'], 'Does not contain ArIII in list'




import numpy

from mangadap.util.extinction import GalacticExtinction


def test_ccm():
    ccm = GalacticExtinction(form='CCM')

    Av = 0.068
    ebv = Av/ccm.rv

    wave = numpy.array([ 3500, 5000, 7000, 10000])

    assert numpy.allclose(ccm.compute(wave, ebv),
                          numpy.array([1.10482203, 1.07281569, 1.04810482, 1.02562548])), \
                'Bad CCM redenning values.'


def test_odonnell():
    ccm = GalacticExtinction(form='CCM')
    odonnell = GalacticExtinction(form='ODonnell')

    Av = 0.068
    ebv = Av/odonnell.rv

    wave = numpy.array([ 3500, 5000, 7000, 10000])

    assert numpy.all(odonnell.compute(wave, ebv)/ccm.compute(wave, ebv) < 1.001), \
            'O\'Donnell values incorrect.'


def test_fitzpatrick():
    ccm = GalacticExtinction(form='CCM')
    fitzpatrick = GalacticExtinction(form='FM')

    Av = 0.068
    ebv = Av/fitzpatrick.rv

    wave = numpy.array([ 3500, 5000, 7000, 10000])

    assert numpy.all(fitzpatrick.compute(wave, ebv)/ccm.compute(wave, ebv) < 1.), \
            'Fitzpatrick values incorrect.'


def test_calzetti():
    ccm = GalacticExtinction(form='CCM')
    calzetti = GalacticExtinction(form='Calzetti')

    Av = 0.068
    ebv = Av/calzetti.rv

    wave = numpy.array([ 3500, 5000, 7000, 10000])

    assert numpy.all(calzetti.compute(wave, ebv)/ccm.compute(wave, ebv) < 1.02), \
            'Calzetti values incorrect.'




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""

Container class for databases of spectral features. This implements
the base class used by :class:`mangadap.par.artifactdb.ArtifactDB`,
:class:`mangadap.par.emissionlinedb.EmissionLineDB`,
:class:`mangadap.par.emissionmomentsdb.EmissionMomentsDB`,
:class:`mangadap.par.absorptionindexdb.AbsorptionIndexDB`, and
:class:`mangadap.par.bandheadindexdb.BandheadIndexDB`.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
from pathlib import Path

from IPython import embed

import numpy

from pydl.pydlutils.yanny import yanny

from ..config import defaults
from .parset import ParDatabase
from ..util.parser import DefaultConfig
from ..proc import util

# TODO: Enable searching of local directory

class SpectralFeatureDB(ParDatabase):
    """
    Basic container class for the parameters databases of spectral
    features.  This is the base class for all of the following:

        - class used by :class:`~mangadap.par.artifactdb.ArtifactDB`
        - :class:`~mangadap.par.emissionlinedb.EmissionLineDB`
        - :class:`~mangadap.par.emissionmomentsdb.EmissionMomentsDB`
        - :class:`~mangadap.par.absorptionindexdb.AbsorptionIndexDB`
        - :class:`~mangadap.par.bandheadindexdb.BandheadIndexDB`

    See :class:`~mangadap.par.parset.ParDatabase` for additional
    attributes.

    Each derived class must define its own default directory that
    contains the relevant databases, the class that defines the base
    :class:`~mangadap.par.parset.ParSet` for each
    :class:`~mangadap.par.parset.ParDatabase` entry, and the method
    that parses the parameter file into the parameter list.

    The primary instantiation requires the SDSS parameter file. To
    instantiate using a keyword (and optionally a directory that
    holds the parameter files), use :func:`from_key`.

    Args:
        parfile (:obj:`str`, `Path`_):
            The SDSS parameter file with the emission-line database.

    Attributes:
        key (:obj:`str`):
            Database signifying keyword
        file (:obj:`str`):
            File with the emission-line data
        size (:obj:`int`):
            Number of elements in the database. 
    """
    default_data_dir = None
    def __init__(self, parfile):
        # TODO: The approach here (read using yanny, set to par
        # individually, then covert back to record array using
        # ParDatabase) is stupid...
        _parfile = Path(parfile).resolve()
        if not _parfile.exists():
            raise FileNotFoundError(f'{_parfile} does not exist!')

        self.key = util.get_database_key(str(_parfile))
        self.file = str(_parfile)
        self.size = None

        # _parse_yanny() has to set `size` for each subclass.
        ParDatabase.__init__(self, self._parse_yanny())

        # Ensure that all indices are unique
        if len(numpy.unique(self.data['index'])) != self.size:
            raise ValueError(f'Database indices for {self.key} are not all unique!')

    def _parse_yanny(self):
        raise NotImplementedError(f'_parse_yanny not defined for {self.__class__.__name__}.')

    @classmethod
    def default_path(cls):
        """
        Return the default path with the emission-line databases.
        """
        if cls.default_data_dir is None:
            raise ValueError(f'Default data directory undefined for {cls.__class__.__name__}.')
        return defaults.dap_data_root() / cls.default_data_dir

    @classmethod
    def available_databases(cls, directory_path=None):
        """
        Return the list of available database files.

        Args:
            directory_path (:obj:`str`, optional):
                 Root path with the database files. If None, uses the
                 default directory defined by :func:`default_path`.

        Returns:
            :obj:`dict`: A dictionary with the database files and
            associated keyword.

        Raises:
            NotADirectoryError:
                Raised if the provided or default directory does not
                exist.
            ValueError:
                Raised if the keywords found for all the ``*.par``
                files are not unique.
        """
        if directory_path is None:
            directory_path = cls.default_path()
        if not directory_path.exists():
            raise NotADirectoryError(f'{directory_path} not found!')
        files = list(directory_path.glob('*.par'))
        keys = [util.get_database_key(str(f)) for f in files]
        if len(keys) != len(numpy.unique(keys)):
            raise ValueError(f'Keys read for par files in {directory_path} are not unique!  Names '
                             'of par files must be case-insensitive and unique.')
        return {k: f for k,f in zip(keys, files)}

    @classmethod
    def from_key(cls, key, directory_path=None):
        r"""
        Instantiate the object using a keyword.

        Args:
            key (:obj:`str`):
                Keyword selecting the database to use.
            directory_path (:obj:`str`, optional):
                Root path with the database parameter files. If None,
                uses the default set by :func:`available_databases`.
                Note that the file search includes *any* file with a
                ``.par`` extension. The root of the file should be
                case-insensitive.
        """
        databases = cls.available_databases(directory_path=directory_path)
        available_keys = list(databases.keys())
        if key not in available_keys:
            raise KeyError(f'No database found with key {key}.  Keywords found '
                           f'are: {available_keys}')
        return cls(databases[key])





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""

A catch-all module with miscellaneous utility functions.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import numpy
#import time
#from scipy import sparse

def line_coeff(p1, p2):
    r"""
    Given two points on a line return the slope and intercept calculated as

    .. math:: 

        m &= \frac{y_1 - y_2}{x_1 - x_2} \\
        b &= y_2 - m\ x_2

    Args:
        p1 (array-like): A two-element :math:`(x_1,y_1)` array with one
            of the two points on the line.
        p2 (array-like): A two-element :math:`(x_2,y_2)` array with one
            of the two points on the line.

    Returns:
        float: The slope (:math:`m`) and intercept (:math:`b`) of the
        line, respectively.

    .. warning:: 
        Performs **no** checks of the input.
    
    """
    m = (p1[1] - p2[1])/(p1[0]-p2[0])
    b = p2[1] - m*p2[0]
    return m, b


#def where_not(indx, size):
#    """
#    Return a tuple with the indices of a vector that were *not* selected
#    by a call to `numpy.where`_.  **The function currently only works
#    for 1D vectors.**
#
#    Args:
#        indx (tuple): Tuple returned by a call to `numpy.where`_ for a
#            1D vector.
#        size (int): Length of the original vector in the call to
#            `numpy.where`_.
#
#    .. warning:: 
#        Performs **no** checks of the input.
#
#    """
#    return (numpy.setdiff1d(numpy.arange(0,size), indx[0]),)
    

#def inverse_with_zeros(v, absolute=True):
#    """
#    Invert an array with zeros, and handle them by setting the inverse
#    to zero.
#
#    Args:
#        v (array-like): Array to invert
#        absolute (bool): Forces that the absolute value must be larger
#            than 0.  Otherwise, any non-positive value is also masked
#            with a zero.
#
#    Returns:
#        numpy.ndarray: Inverse of the array when the vector has values
#        that are greater than 0, otherwise set to 0.0.
#    """
#    if isinstance(v, numpy.ma.MaskedArray):
#        if not absolute:
##            v.mask |= numpy.invert(v > 0)
#            v[numpy.invert(v > 0)] = numpy.ma.masked
#        return 1.0/v
##    _v = numpy.asarray(v).astype(float)
#    _v = numpy.atleast_1d(v).astype(float)
#    indx = numpy.absolute(_v) > 0 if absolute else _v > 0
#    _v[indx] = 1.0/_v[indx]
#    _v[numpy.invert(indx)] = 0.0
#    return _v





#!/usr/bin/env python3

"""
Dynamically build the rst documentation of the bitmasks.
"""

import os
import time
import textwrap
import numpy

#-----------------------------------------------------------------------------

def write_bitmask(bitmask_class, opath, class_link=True):
    ofile = os.path.join(opath, '{0}.rst'.format(bitmask_class.__name__.lower()))
    lines = bitmask_class().to_rst_table(header=False, class_link=class_link)
    with open(ofile, 'w') as f:
        f.write('\n'.join(lines))

if __name__ == '__main__':
    t = time.perf_counter()

    from pkg_resources import resource_filename
    root = os.path.dirname(resource_filename('mangadap', ''))
    path = os.path.join(root, 'docs', 'tables')
    if not os.path.isdir(path):
        os.makedirs(path)

    # Tables to write:
    #
    from mangadap.util.drpbitmask import DRPQuality3DBitMask
    write_bitmask(DRPQuality3DBitMask, path)
    from mangadap.util.drpbitmask import DRPFitsBitMask
    write_bitmask(DRPFitsBitMask, path)

    from mangadap.dapfits import DAPQualityBitMask
    write_bitmask(DAPQualityBitMask, path)
    from mangadap.dapfits import DAPMapsBitMask
    write_bitmask(DAPMapsBitMask, path)
    from mangadap.dapfits import DAPCubeBitMask
    write_bitmask(DAPCubeBitMask, path)

    from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectraBitMask
    write_bitmask(SpatiallyBinnedSpectraBitMask, path, class_link=False)
    from mangadap.proc.templatelibrary import TemplateLibraryBitMask
    write_bitmask(TemplateLibraryBitMask, path, class_link=False)
    from mangadap.proc.stellarcontinuummodel import StellarContinuumModelBitMask
    write_bitmask(StellarContinuumModelBitMask, path, class_link=False)
    from mangadap.proc.emissionlinemoments import EmissionLineMomentsBitMask
    write_bitmask(EmissionLineMomentsBitMask, path, class_link=False)
    from mangadap.proc.emissionlinemodel import EmissionLineModelBitMask
    write_bitmask(EmissionLineModelBitMask, path, class_link=False)
    from mangadap.proc.spectralindices import SpectralIndicesBitMask
    write_bitmask(SpectralIndicesBitMask, path, class_link=False)

    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))



